import datetime
import string
import re
import webbrowser
import pathlib
import urllib.parse
import subprocess
import sys
import os

from pytodotxt import Task, TodoTxt

from pter import common
from pter.searcher import get_relative_date
from pter.source import Source


DATETIME_FMT = '%Y-%m-%d-%H-%M-%S'
FORMAT_TOKEN_RE = re.compile('^([^a-z]*)([a-z][a-z-]*)(.*)$')


def duration_as_str(duration):
    seconds = int(duration.total_seconds())
    hours = seconds // 3600
    minutes = (seconds % 3600) // 60
    
    result = ''
    if hours > 0:
        result += f'{hours}h'
    if minutes > 0:
        result += f'{minutes}m'

    return result


def parse_duration(text):
    duration = datetime.timedelta(0)

    sign = 1
    if text.startswith('-'):
        sign = -1
        text = text[1:]
    elif text.startswith('+'):
        text = text[1:]

    value = ''
    for char in text.lower():
        if char in string.digits:
            value += char
        elif char == 'h' and len(value) > 0:
            duration += datetime.timedelta(hours=int(value))
            value = ''
        elif char == 'm' and len(value) > 0:
            duration += datetime.timedelta(minutes=int(value))
            value = ''
        elif char == 's' and len(value) > 0:
            duration += datetime.timedelta(seconds=int(value))
            value = ''
        elif char == 'd' and len(value) > 0:
            duration += datetime.timedelta(days=int(value))
            value = ''
        elif char == 'w' and len(value) > 0:
            duration += datetime.timedelta(days=int(value)*7)
            value = ''
        else:
            # parse error
            return None

    if len(value) > 0:
        duration += datetime.timedelta(minutes=int(value))
    
    duration *= sign

    return duration


def task_due_in_days(task, default=None):
    if len(task.attr_due) == 0:
        return default
    today = datetime.datetime.now().date()
    try:
        then = datetime.datetime.strptime(task.attr_due[0], Task.DATE_FMT).date()
        return (then - today).days
    except ValueError:
        return default


def dehumanize_dates(text, tags=None):
    """Replace occurrences of relative dates in tags"""
    if tags is None:
        tags = ['due', 't']

    offset = 0
    found_any = True

    while found_any and offset < len(text):
        found_any = False
        for match in Task.KEYVALUE_RE.finditer(text, offset):
            if match is None:
                offset = len(text)
                break
            found_any = True
            if match.group(2) in tags and not Task.DATE_RE.match(match.group(3)):
                then = get_relative_date(match.group(3))
                if then is not None:
                    then = then.strftime(Task.DATE_FMT)
                    text = text[:match.start(3)] + \
                           then + \
                           text[match.end(3):]
                    offset = match.start(3) + len(then)
                    break
            offset = match.end(3) + 1
    return text


def ensure_up_to_date(task):
    ok = True
    if task.todotxt.refresh():
        ok = False
        task.todotxt.parse()
        tasks = [other for other in source.tasks if other.raw.strip() == task.raw.strip()]
        if len(tasks) > 0:
            ok = True
            task = tasks[0]

    if ok:
        return task

    return None


def toggle_tracking(task):
    if 'tracking' in task.attributes:
        return update_spent(task)
    task.add_attribute('tracking', datetime.datetime.now().strftime(DATETIME_FMT))
    return True


def toggle_done(task, clear_contexts):
    task.is_completed = not task.is_completed
    if task.is_completed:
        task.completion_date = datetime.datetime.now().date()
        if task.priority is not None:
            task.add_attribute('pri', task.priority)
            task.priority = None
        if len(clear_contexts) > 0 and task.description is not None:
            for context in clear_contexts:
                while f'@{context}' in task.description:
                    task.remove_context(context)
    else:
        task.completion_date = None
        attrs = task.attributes
        if 'pri' in attrs:
            task.priority = attrs['pri'][0]
            task.remove_attribute('pri')
    task.parse(str(task))


def toggle_hidden(task):
    is_hidden = len(task.attr_h) > 0 and task.attr_h[0] == '1'
    if len(task.attr_h) > 0:
        is_hidden = task.attr_h[0] == '1'
        task.remove_attribute('h', '1')
        if not is_hidden:
            task.add_attribute('h', '1')
    else:
        task.add_attribute('h', '1')


def sign(n):
    if n < 0:
        return -1
    elif n > 0:
        return 1
    return 0


def build_sort_order(order):
    sort_order = []
    for part in order.split(','):
        if part == 'completed':
            sort_order.append(lambda t: t.is_completed)
        elif part == 'due_in':
            sort_order.append(lambda t: task_due_in_days(t, sys.maxsize))
        elif part == 'priority':
            sort_order.append(lambda t: t.priority or 'ZZZ')
        elif part == 'linenr':
            sort_order.append(lambda t: t.linenr)
        elif part == 'file':
            sort_order.append(lambda t: t.todotxt.displayname.lower())
        elif part == 'project':
            sort_order.append(lambda t: sorted(t.projects)[0].lower() if len(t.projects) > 0 else 'zzzzz')
        elif part == 'context':
            sort_order.append(lambda t: sorted(t.contexts)[0].lower() if len(t.contexts) > 0 else 'zzzzz')
    return sort_order


def sort_fnc(a, order):
    if isinstance(a, tuple):
        task, _ = a
    else:
        task = a
    return [fnc(task) for fnc in order]


def update_spent(task):
    now = datetime.datetime.now()
    tracking = task.attr_tracking
    raw_spent = task.attr_spent

    if len(tracking) == 0:
        return False

    try:
        then = datetime.datetime.strptime(tracking[0], DATETIME_FMT)
    except ValueError:
        return False

    if len(raw_spent) > 0:
        spent = parse_duration(raw_spent[0])
        if spent is None:
            self.status_bar.addstr(0, 0, "Failed to parse 'spent' time", self.color(Window.ERROR))
            self.status_bar.noutrefresh()
            return False
    else:
        spent = datetime.timedelta(0)
    
    diff = now - then
    if diff <= datetime.timedelta(minutes=1):
        diff = datetime.timedelta(0)

    task.remove_attribute('tracking', tracking[0])

    # TODO: make the minimal duration configurable
    if diff >= datetime.timedelta(minutes=1):
        spent = duration_as_str(spent + diff)
        if len(raw_spent) == 0:
            task.add_attribute('spent', spent)
        else:
            task.replace_attribute('spent', raw_spent[0], spent)

        return True

    return True


def human_friendly_date(text):
    if isinstance(text, datetime.datetime):
        then = text.date
    elif isinstance(text, datetime.date):
        then = text
    elif isinstance(text, str):
        try:
            then = datetime.datetime.strptime(text, Task.DATE_FMT).date()
        except ValueError:
            return text
    today = datetime.date.today()
    diff = abs((then - today).days)
    then_wd = ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday',
               'friday', 'saturday', 'sunday'][then.isoweekday()]

    if today == then:
        return 'today'
    if diff == 1:
        if then > today:
            return 'tomorrow'
        return 'yesterday'
    if diff == 2:
        if then > today:
            return 'in 2 days'
        return '2 days ago'
    if diff == 3:
        if then > today:
            return 'in 3 days'
    if diff < 8:
        if then > today:
            return 'next ' + then_wd
        return 'last ' + then_wd
    if diff < 34:
        diff_wk = diff//7
        plural = '' if diff_wk == 1 else 's'
        if then > today:
            return f'in {diff_wk} week{plural}'
        return f'{diff_wk} week{plural} ago'
    if diff < 180:
        diff_mnth = diff//30
        plural = '' if diff_mnth == 1 else 's'
        if then > today:
            return f'in {diff_mnth} month{plural}'
        return f'{diff_mnth} month{plural} ago'
    if diff < 350:
        if then > today:
            return 'in half a year'
        return 'half a year ago'
    return then.strftime(Task.DATE_FMT)


def parse_task_format(text):
    rawtokens = []
    token = ''
    in_block = False
    
    for char in text:
        if in_block:
            if char == '}':
                token += char
                rawtokens.append(token)
                token = ''
                in_block = False
                continue
        elif char == '{':
            if len(token) > 0:
                rawtokens.append(token)
            token = char
            in_block = True
            continue
        token += char

    if len(token) > 0:
        rawtokens.append(token)

    tokens = []
    for token in rawtokens:
        if len(token) > 2 and token.startswith('{') and token.endswith('}'):
            align = None
            token = token[1:-1]
            extra_left = None
            extra_right = None
            if ':' in token and len(token) > 3:
                token, align = token.split(':')

            match = FORMAT_TOKEN_RE.match(token)

            if match is None:
                continue

            if len(match.group(1)) > 0:
                extra_left = match.group(1)
            token = match.group(2)
            if len(match.group(3)) > 0:
                extra_right = match.group(3)

            tokens.append((token, align, extra_left, extra_right))
        else:
            tokens.append(token)

    return tokens


def unquote(text):
    if len(text) <= 1:
        return text
    if text[0] in '"\'' and text[0] == text[-1]:
        return text[1:-1]
    return text


def open_manual():
    docloc = common.HERE / "docs" / "pter.html"

    if docloc.exists():
        webbrowser.open('file://' + str(docloc))


def parse_searches():
    if not common.SEARCHES_FILE.exists():
        return {}

    searches = {}
    with open(common.SEARCHES_FILE, 'rt', encoding="utf-8") as fd:
        for linenr, line in enumerate(fd.readlines()):
            if '=' not in line:
                continue
            name, searchdef = line.split("=", 1)
            name = name.strip()
            searchdef = searchdef.strip()
            if len(name) == 0 or len(searchdef) == 0:
                continue
            searches[name.strip()] = searchdef.strip()

    return searches


def parse_templates():
    if not common.TEMPLATES_FILE.exists():
        return {}

    templates = {}
    with open(common.TEMPLATES_FILE, 'rt', encoding="utf-8") as fd:
        for linenr, line in enumerate(fd.readlines()):
            if '=' not in line:
                continue
            name, templatedef = line.split("=", 1)
            name = name.strip()
            templatedef = templatedef.strip()
            if len(name) == 0 or len(templatedef) == 0:
                continue
            templates[name.strip()] = templatedef.strip()

    return templates


def save_searches(searches):
    common.SEARCHES_FILE.parent.mkdir(parents=True, exist_ok=True)
    with open(common.SEARCHES_FILE, 'wt', encoding="utf-8") as fd:
        for name in sorted(searches.keys()):
            value = searches[name].strip()
            if len(value) == 0:
                continue
            fd.write(f"{name} = {value}\n")


def save_templates(templates):
    common.TEMPLATES_FILE.parent.mkdir(parents=True, exist_ok=True)
    with open(common.TEMPLATES_FILE, 'wt', encoding="utf-8") as fd:
        for name in sorted(templates.keys()):
            value = templates[name].strip()
            if len(value) == 0:
                continue
            fd.write(f"{name} = {value}\n")


def update_displaynames(sources):
    if len(sources) == 0:
        return

    pos = 2
    while True:
        displaynames = {source.displayname for source in sources}
        if len(displaynames) == len(sources):
            return
        for source in sources:
            source.displayname = os.sep.join(str(source.filename).split(os.sep)[-1*pos:])
        pos += 1


def execute_delegate_action(task, to_attr, marker, action):
    if action == common.DELEGATE_ACTION_NONE:
        return

    recipient = ''
    if len(to_attr) > 0 and to_attr in task.attributes:
        recipient = ','.join(task.attributes[to_attr])

    if action == common.DELEGATE_ACTION_MAIL:
        # filter out "to:" (to_attr) and the delegation marker
        text = ' '.join([word for word in str(task).split(' ')
                         if word != marker
                            and (len(to_attr) == 0
                                 or not word.startswith(to_attr + ':'))])
        uri = 'mailto:' + urllib.parse.quote(recipient) + '?Subject=' + urllib.parse.quote(text)
        webbrowser.open(uri)


def new_task_id(sources, prefix=""):
    """Generate a new unique task ID
    The task ID will be unique for the given sources, and with the given prefix.
    """
    existing_ids = set()
    for source in sources:
        existing_ids |= {int(key[len(prefix):]) for key in source.task_ids
                         if key.startswith(prefix) and key[len(prefix):].isnumeric()}

    if len(existing_ids) > 0:
        highest = max(existing_ids)
    else:
        highest = 0

    return prefix + str(highest+1)


def query_latest_version():
    installed_re = re.compile(r'^[\s]+IN_STALLED:[\s]+([^\s]+)(.*)$')
    latest_re = re.compile(r'^[\s]+LATEST:[\s]+([^\s]+)$')
    output = subprocess.run([sys.executable, "-m", "pip", "search", common.PROGRAMNAME], capture_output=True)
    for line in str(output.stdout, 'utf-8').split("\n"):
        match = installed_re.match(line)
        if match is not None:
            if len(match.groups()) > 1:
                return match.group(1)
        match = latest_re.match(line)
        if match is not None:
            return match.group(1)

    return ''


def open_sources(args):
    sources = [Source(TodoTxt(pathlib.Path(fn).expanduser().resolve())) for fn in args.filename]
    for source in sources:
        if source.filename.exists():
            source.parse()
    return sources


def create_from_search(searcher):
    text = []
    for group, prefix in [('contexts', '@'), ('projects', '+'), ('words', '')]:
        text += [prefix+word for word in getattr(searcher, group, set())]
    return ' '.join(text)


def auto_task_id(sources, text):
    words = []
    for word in text.split(' '):
        if word.startswith('id:'):
            _, base = word.split(':', 1)
            if '#' in base:
                base, keyword = base.split('#', 1)
                if keyword == '' or keyword == 'auto':
                    word = 'id:{}'.format(new_task_id(sources, base))
        words.append(word)
    return ' '.join(words)




import unittest
import pathlib

from pter import utils


class FakeSource:
    def __init__(self, location):
        self.filename = pathlib.Path(location)
        self.displayname = self.filename.name


class TestDisplayNames(unittest.TestCase):
    def setUp(self):
        pass

    def test_case1(self):
        sources = [FakeSource('/home/me/some/place/todo.txt'),
                   FakeSource('/home/me/other.txt')]
        utils.update_displaynames(sources)

        self.assertEqual(sources[0].displayname, 'todo.txt')
        self.assertEqual(sources[1].displayname, 'other.txt')

    def test_case2(self):
        sources = [FakeSource('/home/me/some/todo.txt')]
        utils.update_displaynames(sources)
        self.assertEqual(sources[0].displayname, 'todo.txt')


    def test_case3(self):
        sources = [FakeSource('/home/me/some/todo.txt'),
                   FakeSource('/home/me/other/todo.txt')]
        utils.update_displaynames(sources)
        self.assertEqual(sources[0].displayname, 'some/todo.txt')
        self.assertEqual(sources[1].displayname, 'other/todo.txt')

    def test_case4(self):
        sources = [FakeSource('/me/some/todo.txt'),
                   FakeSource('/me/todo.txt'),
                   FakeSource('/me/other/todo.txt')]
        utils.update_displaynames(sources)
        self.assertEqual(sources[0].displayname, 'some/todo.txt')
        self.assertEqual(sources[1].displayname, 'me/todo.txt')
        self.assertEqual(sources[2].displayname, 'other/todo.txt')




from pter import common


class Source:
    def __init__(self, source):
        self.source = source
        self.last_change = 0
        self.displayname = self.filename.name
        self._contexts = set()
        self._projects = set()
        self._task_by_id = {}
        self.refresh()

    @property
    def contexts(self):
        return self._contexts

    @property
    def projects(self):
        return self._projects

    @property
    def task_ids(self):
        return set(self._task_by_id.keys())

    def __lt__(self, other):
        return self.displayname < other.displayname

    def update_contexts_and_projects(self):
        self._projects = set()
        self._contexts = set()
        self._task_by_id = {}
        for task in self.source.tasks:
            self.update_from_task(task)

    def update_from_task(self, task):
        if task.description is None or len(task.description) == 0:
            return
        for word in task.description.split(' '):
            if word.startswith('+'):
                self._projects.add(word)
            elif word.startswith('@'):
                self._contexts.add(word)
            elif word.startswith(common.ATTR_ID + ':'):
                _, taskid = word.split(':', 1)
                if len(taskid) > 0:
                    self._task_by_id[taskid] = task

    def refresh(self):
        last_change = self.last_change
        if self.filename.exists():
            last_change = self.source.filename.stat().st_mtime
        has_changed = False

        if last_change != self.last_change:
            self.last_change = last_change
            has_changed = True

        return has_changed

    def save(self, safe=True):
        self.source.save(safe=safe)
        self.refresh()

    def parse(self):
        self._projects = set()
        self._contexts = set()
        self._task_by_id = {}
        for task in self.source.parse():
            task.todotxt = self
            self.update_from_task(task)
        return self.source.tasks

    def task(self, taskid, default=None):
        return self._task_by_id.get(taskid, default)

    @property
    def tasks(self):
        return self.source.tasks

    @property
    def filename(self):
        return self.source.filename




import datetime
import unittest

from pytodotxt import Task
from pter import utils


class TestPter(unittest.TestCase):
    def test_parse_duration(self):
        text = '1h10m'
        then = utils.parse_duration(text)
        self.assertEqual(then, datetime.timedelta(hours=1, minutes=10))

    def test_update_spent(self):
        then = (datetime.datetime.now() - datetime.timedelta(minutes=10)).strftime(utils.DATETIME_FMT)
        task = Task(f'Something spent:1h10m tracking:{then}')

        self.assertTrue(utils.update_spent(task))
        self.assertEqual(task.attributes, {'spent': ['1h20m']})


class TestHumanFriendlyDates(unittest.TestCase):
    def today(self, daydiff=0):
        return (datetime.date.today() + datetime.timedelta(days=daydiff)).strftime(Task.DATE_FMT)

    def test_today(self):
        then = self.today()
        self.assertEqual(utils.human_friendly_date(then), "today")

    def test_tomorrow(self):
        then = self.today(+1)
        self.assertEqual(utils.human_friendly_date(then), "tomorrow")

    def test_yesterday(self):
        then = self.today(-1)
        self.assertEqual(utils.human_friendly_date(then), "yesterday")


if __name__ == '__main__':
    unittest.main()




import unittest
import pathlib

from pytodotxt import Task

from pter import utils
from pter.source import Source


class FakeSource:
    def __init__(self, tasks):
        self.tasks = tasks
        self.filename = pathlib.Path('/tmp/test.txt')


class TestTaskIDGeneration(unittest.TestCase):
    def setUp(self):
        self.sources = [
                Source(FakeSource([Task('task id:30'), Task('task id:prefixed22')])),
                Source(FakeSource([Task('task id:41')]))]
        self.sources[0].update_contexts_and_projects()
        self.sources[1].update_contexts_and_projects()

    def test_no_prefix_no_existing(self):
        sources = [Source(FakeSource([]))]
        sources[0].update_contexts_and_projects()
        nextid = utils.new_task_id(sources)
        self.assertEqual(nextid, '1')

    def test_no_prefix_existing(self):
        nextid = utils.new_task_id(self.sources)
        self.assertEqual(nextid, '42')

    def test_prefix_no_existing(self):
        nextid = utils.new_task_id(self.sources, 'project')
        self.assertEqual(nextid, 'project1')

    def test_prefix_existing(self):
        nextid = utils.new_task_id(self.sources, 'prefixed')
        self.assertEqual(nextid, 'prefixed23')

    def test_decimals(self):
        sources = [Source(FakeSource([Task('task id:9'), Task('task id:12')]))]
        sources[0].update_contexts_and_projects()
        nextid = utils.new_task_id(sources)
        self.assertEqual(nextid, '13')




import argparse
import locale
import pathlib
import os
import sys
import datetime
import configparser
import string
import webbrowser
import curses
import curses.panel

from pytodotxt import TodoTxt, Task, version

from pter import common
from pter import utils
from pter import configuration
from pter.searcher import Searcher, get_relative_date
from pter.key import Key
from pter.source import Source
from pter.tr import tr


class Color:
    def __init__(self, fg, bg=None):
        self.fg = fg
        self.bg = bg

    def pair(self):
        return [self.fg, self.bg]

    def __eq__(self, other):
        return self.fg == other.fg and self.bg == other.bg


class TaskLineGroup:
    def __init__(self):
        self.elements = []

    def append(self, *args, **kwargs):
        self.elements.append(TaskLineElement(*args, **kwargs))


class TaskLine:
    def __init__(self, task, source):
        self.elements = {}
        self.task = task
        self.source = source
        self._due = None

    def add(self, kind, *args, **kwargs):
        self.elements[kind] = TaskLineElement(*args, **kwargs)
        self.elements[kind].name = kind

    @property
    def due(self):
        due = self.task.attr_due
        if self._due is None and len(due) > 0:
            try:
                self._due = datetime.datetime.strptime(due[0], Task.DATE_FMT).date()
            except ValueError:
                self._due = None
        return self._due

    @property
    def is_overdue(self):
        return self.due is not None and self.due < datetime.date.today() and not self.task.is_completed

    @property
    def is_due_tomorrow(self):
        return self.due is not None and self.due == datetime.date.today() + datetime.timedelta(days=1)

    @property
    def is_due_today(self):
        return self.due is not None and self.due == datetime.date.today()


class TaskLineDescription(TaskLineGroup):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = common.TF_DESCRIPTION

    @property
    def space_around(self):
        return True


class TaskLineElement:
    def __init__(self, content, color=None, space_around=False, name=None):
        self.content = content
        self.color = color
        self.space_around = space_around
        self.name = name


class TaskLineSelectionIcon(TaskLineElement):
    def __init__(self, content):
        super().__init__(content, space_around=False)
        self.name = common.TF_SELECTION


class Window:
    def __init__(self, sources, conf):
        self.scr = curses.initscr()
        self.dim = self.scr.getmaxyx()
        self.quit = False
        self.sources = sources
        self.conf = conf

        self.safe_save = conf.bool(common.SETTING_GROUP_GENERAL,
                                   common.SETTING_SAFE_SAVE, 'y')

        self.key_mapping = {'q': 'quit',
                            '^C': 'cancel',
                            '<escape>': 'cancel',
                            '<down>': 'next-item',
                            '<up>': 'prev-item',
                            'j': 'next-item',
                            'k': 'prev-item',
                            '<return>': 'select-item',
                            '<pgup>': 'half-page-up',
                            '<pgdn>': 'half-page-down',
                            '<home>': 'first-item',
                            '<end>': 'last-item',
                            'h': 'toggle-hidden',
                            'd': 'toggle-done',
                            'e': 'edit-task',
                            'n': 'create-task',
                            ':': 'jump-to',
                            '/': 'search',
                            'c': 'search-context',
                            'p': 'search-project',
                            't': 'toggle-tracking',
                            '^L': 'refresh-screen',
                            '^R': 'reload-tasks',
                            'u': 'open-url',
                            '>': 'delegate',
                            'L': 'load-template',
                            'S': 'save-template',
                            'l': 'load-search',
                            's': 'save-search',
                            '?': 'show-help',
                            'm': 'open-manual',}
        self.editor_key_mapping = {'^C': 'cancel',
                                   '<escape>': 'cancel',
                                   '<left>': 'go-left',
                                   '<right>': 'go-right',
                                   '^U': 'del-to-bol',
                                   '<backspace>': 'del-left',
                                   '<del>': 'del-right',
                                   '<home>': 'go-bol',
                                   '<end>': 'go-eol',
                                   '<f6>': 'select-file',
                                   '<return>': 'submit-input',
                                   }
        self.scroll_margin = conf.number(common.SETTING_GROUP_GENERAL,
                                         common.SETTING_SCROLL_MARGIN)
        self.show_numbers = conf.bool(common.SETTING_GROUP_GENERAL,
                                      common.SETTING_SHOW_NUMBERS)
        self.use_colors = conf.bool(common.SETTING_GROUP_GENERAL,
                                    common.SETTING_USE_COLORS)
        self.human_friendly_dates = conf.list(common.SETTING_GROUP_GENERAL,
                                              common.SETTING_HUMAN_DATES)
        self.task_format = utils.parse_task_format(conf.get(common.SETTING_GROUP_GENERAL,
                                                            common.SETTING_TASK_FORMAT,
                                                            common.DEFAULT_TASK_FORMAT))
        self.clear_contexts = [context for context in conf.list(common.SETTING_GROUP_GENERAL,
                                                                common.SETTING_CLEAR_CONTEXT)
                               if len(context) > 0]
        self.add_creation_date = conf.bool(common.SETTING_GROUP_GENERAL, common.SETTING_ADD_CREATED)
        self.create_from_search = conf.bool(common.SETTING_GROUP_GENERAL, common.SETTING_CREATE_FROM_SEARCH)
        self.auto_id = conf.bool(common.SETTING_GROUP_GENERAL, common.SETTING_AUTO_ID)
        self.protos = conf.list(common.SETTING_GROUP_GENERAL,
                                common.SETTING_PROTOCOLS)

        self.selection_indicator = utils.unquote(conf.get(common.SETTING_GROUP_SYMBOLS,
                                                          common.SETTING_ICON_SELECTION))
        self.done_marker = (utils.unquote(conf.get(common.SETTING_GROUP_SYMBOLS,
                                                   common.SETTING_ICON_NOT_DONE)),
                            utils.unquote(conf.get(common.SETTING_GROUP_SYMBOLS,
                                                   common.SETTING_ICON_DONE)))
        self.overflow_marker = (utils.unquote(conf.get(common.SETTING_GROUP_SYMBOLS,
                                                       common.SETTING_ICON_OVERFLOW_LEFT)),
                                utils.unquote(conf.get(common.SETTING_GROUP_SYMBOLS,
                                                       common.SETTING_ICON_OVERFLOW_RIGHT)))
        self.due_marker = (utils.unquote(conf.get(common.SETTING_GROUP_SYMBOLS,
                                                  common.SETTING_ICON_OVERDUE)),
                           utils.unquote(conf.get(common.SETTING_GROUP_SYMBOLS,
                                                  common.SETTING_ICON_DUE_TODAY)),
                           utils.unquote(conf.get(common.SETTING_GROUP_SYMBOLS,
                                                  common.SETTING_ICON_DUE_TOMORROW)))
        self.tracking_marker = utils.unquote(conf.get(common.SETTING_GROUP_SYMBOLS,
                                                      common.SETTING_ICON_TRACKING))

        # mapable functions
        self.functions = {'quit': self.do_quit,
                          'nop': lambda: True,
                          'next-item': self.do_next_task,
                          'prev-item': self.do_prev_task,
                          'half-page-up': self.do_half_page_up,
                          'half-page-down': self.do_half_page_down,
                          'refresh-screen': self.do_refresh_screen,
                          'reload-tasks': self.do_reload_tasks,
                          'search': self.do_start_search,
                          'search-context': self.do_search_context,
                          'search-project': self.do_search_project,
                          'load-template': self.do_load_template,
                          'save-template': self.do_save_template,
                          'load-search': self.do_load_search,
                          'save-search': self.do_save_search,
                          'first-item': self.do_go_first_task,
                          'last-item': self.do_go_last_task,
                          'edit-task': self.do_edit_task,
                          'create-task': self.do_create_task,
                          'jump-to': self.do_jump_to,
                          'open-url': self.do_open_url,
                          'toggle-tracking': self.do_toggle_tracking,
                          'toggle-done': self.do_toggle_done,
                          'toggle-hidden': self.do_toggle_hidden,
                          'show-help': self.do_show_help,
                          'delegate': self.do_delegate,
                          'open-manual': utils.open_manual,}
        self.short_name = {'quit': 'Quit',
                           'cancel': 'Cancel',
                           'select-item': 'Select',
                           'next-item': 'Next item',
                           'prev-item': 'Previous item',
                           'half-page-up': 'Half page up',
                           'half-page-down': 'Half page down',
                           'search': 'Search',
                           'open-url': 'Open URL',
                           'load-template': 'Load task template',
                           'save-template': 'Save task as template',
                           'load-search': 'Load search',
                           'save-search': 'Save search',
                           'search-context': 'Search for context of this task',
                           'search-project': 'Search for project of this task',
                           'first-item': 'First item',
                           'last-item': 'Last item',
                           'edit-task': 'Edit task',
                           'create-task': 'New task',
                           'jump-to': 'Jump to item',
                           'toggle-hidden': 'Set/unset hidden',
                           'toggle-done': 'Set/unset done',
                           'toggle-tracking': 'Start/stop tracking',
                           'show-help': 'Help',
                           'open-manual': 'Read the manual',
                           'go-left': 'Go one character to the left',
                           'go-right': 'Go one character to the right',
                           'go-bol': 'Go to start of line',
                           'go-eol': 'Go to end of line',
                           'del-left': 'Delete to the left',
                           'del-right': 'Delete to the right',
                           'del-to-bol': 'Delete to start of line',
                           'submit-input': 'Apply changes',
                           'select-file': 'Select file',
                           'delegate': 'Delegate task',
                           'refresh-screen': 'Refresh screen',
                           'reload-tasks': 'Reload todo files',
                           }

        curses.start_color()
        self.colors = {}
        self.color_cache = {}
        if curses.has_colors() and self.use_colors:
            curses.use_default_colors()
            self.update_color_pairs()
        elif len(self.selection_indicator) == 0:
            self.selection_indicator = '>'
        curses.noecho()
        curses.cbreak()
        self.scr.keypad(True)
        try:
            curses.curs_set(0)
        except:
            pass

        self.tasks = []
        self.filtered_tasks = []
        self.task_lines = []
        self.max_widths = {}
        self.selected_task = -1
        self.selected_template = None
        self.scroll_offset = 0
        self.must_repaint = True
        self.sort_order = utils.build_sort_order(common.DEFAULT_SORT_ORDER)
        self.sort_order_txt = common.DEFAULT_SORT_ORDER
        self.search = Searcher('',
                               self.conf.bool(common.SETTING_GROUP_GENERAL,
                                              common.SETTING_SEARCH_CASE_SENSITIVE),
                               self.conf.get(common.SETTING_GROUP_GENERAL,
                                             common.SETTING_DEFAULT_THRESHOLD),
                               self.conf.bool(common.SETTING_GROUP_GENERAL,
                                              common.SETTING_HIDE_SEQUENTIAL))
        self.search.update_sources(self.sources)
        self.delegation_marker = self.conf.get(common.SETTING_GROUP_GENERAL,
                                               common.SETTING_DELEG_MARKER)
        self.delegation_action = self.conf.get(common.SETTING_GROUP_GENERAL,
                                               common.SETTING_DELEG_ACTION).lower()
        self.delegate_to = self.conf.get(common.SETTING_GROUP_GENERAL,
                                         common.SETTING_DELEG_TO).lower()
        if self.delegation_action not in common.DELEGATE_ACTIONS:
            self.delegation_action = common.DELEGATE_ACTION_NONE

        for item in self.conf[common.SETTING_GROUP_KEYS]:
            fnc = self.conf.get(common.SETTING_GROUP_KEYS, item, None)
            if fnc is None or fnc not in self.functions:
                continue

            if len(item) == 1:
                self.key_mapping[item] = fnc
            elif len(item) == 2 and item[0] == '^':
                self.key_mapping[item.upper()] = fnc
            elif item in Key.SPECIAL:
                self.key_mapping[item] = fnc
            else:
                raise RuntimeError(f'Invalid key: {item}')
        to_exit = [k for k, fnc in self.key_mapping.items() if fnc == 'quit']
        if len(to_exit) == 0:
            raise RuntimeError("No key to exit")

        for item in self.conf[common.SETTING_GROUP_EDITORKEYS]:
            fnc = self.conf.get(common.SETTING_GROUP_EDITORKEYS, item, None)
            
            if len(item) == 1:
                self.editor_key_mapping[item] = fnc
            elif len(item) == 2 and item[0] == '^':
                self.editor_key_mapping[item.upper()] = fnc
            elif item in Key.SPECIAL:
                self.editor_key_mapping[item] = fnc
            else:
                raise RuntimError(f"Invalid key: {item}")
        to_exit = [k for k, fnc in self.editor_key_mapping.items() if fnc == 'cancel']
        if len(to_exit) == 0:
            raise RuntimeError("No key to cancel editing")
        list_height = self.dim[0]-4
        if len(self.filtered_tasks) > 0:
            self.selected_task = max(0, min(self.selected_task, len(self.filtered_tasks)-1))

        if self.selected_task <= self.scroll_offset + self.scroll_margin:
            self.scroll_offset = max(0, self.selected_task - self.scroll_margin)
        elif self.selected_task >= self.scroll_offset + list_height - self.scroll_margin:
            self.scroll_offset = min(max(0, len(self.filtered_tasks) - list_height),
                                     self.selected_task - list_height + self.scroll_margin + 1)

    def update_color_pairs(self):
        self.colors = {common.SETTING_COL_NORMAL: [Color(7, -1), Color(0, 7)],
                       common.SETTING_COL_INACTIVE: [Color(8), None],
                       common.SETTING_COL_ERROR: [Color(1), None],
                       common.SETTING_COL_PRI_A: [Color(1), None],
                       common.SETTING_COL_PRI_B: [Color(3), None],
                       common.SETTING_COL_PRI_C: [Color(6), None],
                       common.SETTING_COL_CONTEXT: [Color(4), None],
                       common.SETTING_COL_PROJECT: [Color(2), None],
                       common.SETTING_COL_HELP_TEXT: [Color(11, 8), None],
                       common.SETTING_COL_HELP_KEY: [Color(2, 8), None],
                       common.SETTING_COL_OVERFLOW: [Color(11), None],
                       common.SETTING_COL_OVERDUE: [Color(7, 1), Color(1, 7)],
                       common.SETTING_COL_DUE_TODAY: [Color(4), None],
                       common.SETTING_COL_DUE_TOMORROW: [Color(6), None],
                       common.SETTING_COL_TRACKING: [Color(7, 2), Color(2, 7)],
                       }
        if curses.has_colors() and self.use_colors:
            for colorname in self.conf[common.SETTING_GROUP_COLORS]:
                colpair = self.conf.color_pair(common.SETTING_GROUP_COLORS, colorname)

                if colpair is None:
                    continue
                fg, bg = colpair
                pairidx = 0

                if colorname.startswith('sel-'):
                    colorname = colorname[4:]
                    pairidx = 1

                if colorname not in self.colors:
                    # TODO: log the error in the configuration
                    continue

                self.colors[colorname][pairidx] = Color(fg, bg)

            # register the color pairs
            for colorname in self.colors.keys():
                self.color(colorname, 0)
                self.color(colorname, 1)

            for number, key in enumerate(self.conf[common.SETTING_GROUP_HIGHLIGHT]):
                hlcol = Color(*self.conf.color_pair(common.SETTING_GROUP_HIGHLIGHT, key))

                variant = 0
                if key.startswith('sel-'):
                    variant = 1
                    key = key[4:]

                if len(key.strip()) == 0:
                    continue

                if 'hl:' + key not in self.colors:
                    self.colors['hl:' + key] = [None, None]

                self.colors['hl:' + key][variant] = hlcol

                # and initialize the color
                self.color('hl:' + key, variant)

    def color(self, colorname, variant=0, default=None):
        """Return a color pair number for use with curses attributes
        variant can be 0/False (for normal) or 1/True (for selected text)
        If the color pair does not exist yet, it is registered, if possible."""

        if colorname is None:
            colorname = default or common.SETTING_COL_NORMAL

        if not self.use_colors or colorname not in self.colors:
            return 0

        if variant is True:
            variant = 1
        if variant is False:
            variant = 0
        if default is None:
            default = common.SETTING_COL_NORMAL

        colors = self.colors[colorname]
        if variant >= len(colors):
            raise ValueError(variant)

        color = colors[variant]
        if color is None:
            if variant > 0 and colors[0] is not None:
                color = colors[0].pair()
            else:
                color = self.colors[default][variant].pair()
        else:
            color = color.pair()

        default_variant = self.colors[default][variant]
        if default_variant is None and variant > 0:
            default_variant = self.colors[default][0]
        if default_variant is None:
            default_variant = self.colors[common.SETTING_COL_NORMAL][variant]

        if color[0] is None:
            color[0] = default_variant.fg
        if color[1] is None:
            color[1] = default_variant.bg
            if color[1] is None and self.colors[default][0] is not None:
                color[1] = self.colors[default][0].bg
            if color[1] is None:
                color[1] = self.colors[common.SETTING_COL_NORMAL][variant].bg

        color = (color[0], color[1])

        if color in self.color_cache:
            return curses.color_pair(self.color_cache[color])

        next_id = len(self.color_cache) + 1
        if next_id >= curses.COLOR_PAIRS:
            # sucks, we ran out of numbers
            return 0

        try:
            curses.init_pair(next_id, *color)
            self.color_cache[color] = next_id
        except curses.error:
            self.color_cache[color] = 0
            return 0
        return next_id

    def build_screen(self):
        self.scr.attrset(self.color(common.SETTING_COL_NORMAL))
        self.scr.erase()
        self.scr.noutrefresh()
        self.paint_search_bar()
        self.paint_status_bar()
        self.paint_shortcut_bar()
        self.scr.addstr(1, 0, '─'*self.dim[1])
        self.scr.noutrefresh()

    def paint_search_bar(self):
        self.scr.move(0, 0)
        self.scr.clrtoeol()
        attrs = 0
        text = self.search.text.strip()
        if len(text) == 0:
            text = tr('(no search active)')
            attrs = self.color(common.SETTING_COL_INACTIVE)
        self.scr.addstr(0, 1, text, attrs)
        self.scr.noutrefresh()

    def paint_status_bar(self, text="", attrs=None):
        y = self.dim[0]-2
        self.scr.move(y, 0)
        self.scr.clrtoeol()
        if attrs is None:
            attrs = self.color(common.SETTING_COL_NORMAL)
        if len(text) > 0:
            self.scr.addstr(y, 0, text, attrs)
        self.scr.noutrefresh()

    def paint_shortcut_bar(self, actions=None, mapping=None):
        if mapping is None:
            mapping = self.key_mapping
        if actions is None:
            actions = ['show-help', 'quit', 'edit-task',
                       'create-task', 'search', 'load-search', 'save-search',
                       'toggle-done', 'jump-to', 'next-item', 'prev-item']

        y = self.dim[0]-1
        self.scr.move(y, 0)
        self.scr.clrtoeol()
        self.scr.noutrefresh()

        x = 1
        for action in actions:
            label = self.short_name.get(action, None)
            if label is None:
                continue
            label = tr(label) + " "

            keys = [k for k, v in mapping.items() if v == action]
            if len(keys) == 0:
                continue

            if keys[0] in Key.SPECIAL:
                keytext = f" {tr(Key.SPECIAL[keys[0]])} "
            else:
                keytext = f" {tr(keys[0])} "
            if x + len(keytext) + len(label) >= self.dim[1]:
                break

            self.scr.addstr(y, x, keytext, self.color(common.SETTING_COL_HELP_KEY))
            x += len(keytext)
            self.scr.addstr(y, x, label, self.color(common.SETTING_COL_HELP_TEXT))
            x += len(label) + 1
        self.scr.noutrefresh()

    def update_sorting(self):
        new_sort_order = common.DEFAULT_SORT_ORDER
        for part in self.search.text.split(' '):
            if part.startswith('sort:'):
                new_sort_order = part.split(':', 1)[1]
                break
        if self.sort_order_txt != new_sort_order:
            self.sort_order_txt = new_sort_order
            self.sort_order = utils.build_sort_order(new_sort_order)
        self.tasks.sort(key=lambda t: utils.sort_fnc(t, self.sort_order))

    def update_tasks(self):
        self.tasks = []
        for source in self.sources:
            self.tasks += [(task, source) for task in source.tasks]

        self.apply_search()

    def apply_search(self):
        self.update_sorting()
        self.filtered_tasks = [(task, source) for task, source in self.tasks
                               if self.search.match(task)]
        self.must_repaint = True
        self.rebuild_task_lines()

    def rebuild_task_lines(self):
        self.task_lines = []
        self.max_widths = {}
        today = datetime.date.today()

        def update_max_width(name, textlen):
            if name not in self.max_widths:
                self.max_widths[name] = 0
            if isinstance(textlen, str):
                textlen = len(textlen)
            for tf in self.task_format:
                if not isinstance(tf, tuple):
                    continue
                tname, _, left, right = tf
                if tname == name:
                    textlen += 0 if left is None else len(left)
                    textlen += 0 if right is None else len(right)
            self.max_widths[name] = max(textlen, self.max_widths[name])

        update_max_width(common.TF_DONE, max(len(self.done_marker[0]), len(self.done_marker[1])))
        update_max_width(common.TF_SELECTION, self.selection_indicator)
        update_max_width(common.TF_TRACKING, self.tracking_marker)
        update_max_width(common.TF_DUE, max([len(m) for m in self.due_marker]))

        for nr, pair in enumerate(self.filtered_tasks):
            task, source = pair
            line = TaskLine(task, task.todotxt)

            # Selection indicator
            if len(self.selection_indicator) > 0:
                line.elements[common.TF_SELECTION] = TaskLineSelectionIcon(self.selection_indicator)

            # Item number
            text = str(nr+1)
            line.add(common.TF_NUMBER, text)
            update_max_width(common.TF_NUMBER, text)

            # Done marker
            line.add(common.TF_DONE, self.done_marker[1] if task.is_completed else self.done_marker[0])

            # Age
            if task.creation_date is not None:
                text = str((datetime.date.today() - task.creation_date).days)
                line.add(common.TF_AGE, text)
                update_max_width(common.TF_AGE, text)

            # Creation date
            if task.creation_date is not None:
                text = self.date_as_str(task.creation_date, common.TF_CREATED)
                line.add(common.TF_CREATED, text)
                update_max_width(common.TF_CREATED, text)

            # Completion date
            if task.completion_date is not None:
                text = self.date_as_str(task.completion_date, common.TF_COMPLETED)
                line.add(common.TF_COMPLETED, text)
                update_max_width(common.TF_COMPLETED, text)

            # Tracking
            if task.attributes.get(common.ATTR_TRACKING, None) is not None:
                line.add(common.TF_TRACKING, self.tracking_marker)

            # Due marker
            if line.due is not None:
                duedays = str((line.due - today).days)
                line.add(common.TF_DUEDAYS, duedays)
                update_max_width(common.TF_DUEDAYS, duedays)

                if not task.is_completed:
                    if line.due < today:
                        line.add(common.TF_DUE, self.due_marker[0], common.SETTING_COL_OVERDUE)
                    elif line.due == today:
                        line.add(common.TF_DUE, self.due_marker[1], common.SETTING_COL_DUE_TODAY)
                    elif line.due == today + datetime.timedelta(days=1):
                        line.add(common.TF_DUE, self.due_marker[2], common.SETTING_COL_DUE_TOMORROW)

            # Priority marker
            if task.priority is not None:
                pri = task.priority.upper()
                attrs = None
                if pri == 'A':
                    attrs = common.SETTING_COL_PRI_A
                elif pri == 'B':
                    attrs = common.SETTING_COL_PRI_B
                elif pri == 'C':
                    attrs = common.SETTING_COL_PRI_C
                line.add(common.TF_PRIORITY, f"{pri}", attrs)
                update_max_width(common.TF_PRIORITY, '(A)')

            # Description
            description = TaskLineDescription()
            if task.description is not None:
                for word in task.description.split(' '):
                    if len(word) == 0:
                        continue

                    attr = None

                    if word.startswith('@') and len(word) > 1:
                        attr = common.SETTING_COL_CONTEXT
                    elif word.startswith('+') and len(word) > 1:
                        attr = common.SETTING_COL_PROJECT
                    elif ':' in word:
                        key, value = word.split(':', 1)
                        if 'hl:' + key in self.colors:
                            attr = 'hl:' + key
                        if key in [common.ATTR_T, common.ATTR_DUE]:
                            word = key + ':' + self.date_as_str(value, key)
                        if key == common.ATTR_PRI:
                            attr = None
                            value = value.upper()
                            if value == 'A':
                                attr = common.SETTING_COL_PRI_A
                            elif value == 'B':
                                attr = common.SETTING_COL_PRI_B
                            elif value == 'C':
                                attr = common.SETTING_COL_PRI_C
                    description.append(word, attr, space_around=True)
            line.elements[common.TF_DESCRIPTION] = description
            update_max_width(common.TF_DESCRIPTION, ' '.join([e.content for e in description.elements]))

            self.task_lines.append(line)

    def paint_tasks(self, tasks=None):
        """Repaint all tasks on screen.
        If `tasks` is provided, only repaint the tasks with these numbers."""
        list_height = self.dim[0]-4
        y = 2
        nr = 0
        for nr, line in enumerate(self.task_lines):
            if nr < self.scroll_offset:
                continue
            if nr >= self.scroll_offset + list_height:
                break
            if tasks is not None and nr not in tasks:
                continue
            self.write_task(y+nr-self.scroll_offset, 0, nr, line)

        if tasks is None:
            if len(self.filtered_tasks) == 0:
                nr -= 1
            while nr - self.scroll_offset + y <= list_height:
                nr += 1
                self.scr.move(y + nr - self.scroll_offset, 0)
                self.scr.clrtoeol()
        self.scr.noutrefresh()

    def date_as_str(self, text, hint=''):
        if common.TF_ALL in self.human_friendly_dates or hint in self.human_friendly_dates:
            return utils.human_friendly_date(text)
        if not isinstance(text, str):
            return text.strftime(Task.DATE_FMT)
        return text

    def write_task(self, y, x, nr, taskline):
        line = ''
        is_selected = nr == self.selected_task
        is_tracked = len(taskline.task.attr_tracking) > 0

        baseattrs = common.SETTING_COL_NORMAL
        if taskline.is_overdue:
            baseattrs = common.SETTING_COL_OVERDUE
        elif taskline.is_due_tomorrow:
            baseattrs = common.SETTING_COL_DUE_TOMORROW
        elif taskline.is_due_today:
            baseattrs = common.SETTING_COL_DUE_TODAY
        if is_tracked:
            baseattrs = common.SETTING_COL_TRACKING

        maxwidth = self.dim[1]

        def print_element(y, x, maxwidth, element, align, extra):
            cut_off = False
            if isinstance(element, TaskLineSelectionIcon) and not is_selected:
                elem = ''
            elif isinstance(element, TaskLineGroup):
                return print_group(y, x, maxwidth, element, align, extra)
            elif isinstance(element, str):
                element = TaskLineElement(element)
                elem = element.content
            else:
                elem = element.content

            if align is not None:
                width = self.max_widths.get(element.name, 0)
                if extra is not None and extra[0] is not None:
                    width -= len(extra[0])
                if extra is not None and extra[1] is not None:
                    width -= len(extra[1])
                width = max(0, width)
                elem = f"{elem:{align}{width}}"
            elif len(elem) == 0:
                return ''

            if extra_left is not None:
                elem = extra_left + elem

            if extra_right is not None:
                elem = elem + extra_right

            elemlen = len(elem)
            if elemlen >= maxwidth:
                cut_off = True
                elem = elem[:maxwidth]

            self.scr.addstr(y, x, elem, self.color(element.color, is_selected, baseattrs))

            if cut_off:
                attrs = common.SETTING_COL_OVERFLOW
                self.scr.addstr(y, x+maxwidth-len(self.overflow_marker[1]),
                                self.overflow_marker[1],
                                self.color(attrs, is_selected, baseattrs))
            return elem

        def print_group(y, x, maxwidth, group, align, extra):
            line = ''

            if extra is not None and extra[0] is not None:
                self.scr.addstr(y, x + len(line), extra[0], self.color(baseattrs, is_selected))
                line = extra[0]

            if align is not None and align.endswith('>'):
                spacing = align[0] * min(maxwidth - len(line), self.max_widths[group.name] - len(' '.join([e.content for e in group.elements])))
                self.scr.addstr(y, x + len(line), spacing, self.color(baseattrs, is_selected))
                line += spacing

            cut_off = False
            for elnr, element in enumerate(group.elements):
                word = ''

                if elnr > 0 and (element.space_around or group.elements[elnr-1].space_around):
                    if len(line) + 1 >= maxwidth:
                        cut_off = True
                    else:
                        self.scr.addstr(y, x + len(line), " ", self.color(baseattrs, is_selected))
                        line += " "

                cut_off = cut_off or len(line) >= maxwidth

                if not cut_off:
                    if isinstance(element, TaskLineGroup):
                        word = print_group(y, x + len(line), maxwidth-len(line), element, None, None)
                    else:
                        word = print_element(y, x + len(line), maxwidth-len(line), element, None, None)
                    line += word

                if cut_off or len(line) >= maxwidth:
                    break

            if not cut_off and align is not None and align.endswith('<'):
                if len(line) < self.max_widths[group.name]:
                    spacing = align[0] * min(maxwidth-len(line), self.max_widths[group.name]-len(line))
                    self.scr.addstr(y, x + len(line), spacing, self.color(baseattrs, is_selected))
                    line += spacing

            if not cut_off and extra is not None and extra[1] is not None and len(line) + 1 < maxwidth:
                self.scr.addstr(y, x + len(line), extra[1], self.color(baseattrs, is_selected))
            return line

        self.scr.move(y, x)
        self.scr.clrtoeol()
        self.scr.noutrefresh()

        for token in self.task_format:
            align = None
            extra = None
            if isinstance(token, tuple):
                token, align, extra_left, extra_right = token
                extra = (extra_left, extra_right)
                token = taskline.elements.get(token, TaskLineElement('', name=token))
            elif isinstance(token, str):
                token = TaskLineElement(token)
            else:
                raise TypeError()

            x += len(print_element(y, x, maxwidth-x, token, align, extra))

            if x >= maxwidth:
                break

        if x < maxwidth:
            self.scr.addstr(y, x, ' '*(maxwidth-x), self.color(baseattrs, is_selected))

    def recalculate_scrolling(self):
        selected_task = self.selected_task
        scroll_offset = self.scroll_offset

        list_height = self.dim[0]-4
        if len(self.filtered_tasks) == 0:
            selected_task = None
            scroll_offset = 0
        else:
            if selected_task is None:
                selected_task = 0
            selected_task = max(0, min(selected_task, len(self.filtered_tasks)-1))

            if selected_task <= scroll_offset + self.scroll_margin:
                scroll_offset = max(0, selected_task - self.scroll_margin)
            elif selected_task >= scroll_offset + list_height - self.scroll_margin:
                scroll_offset = min(max(0, len(self.filtered_tasks) - list_height),
                                         selected_task - list_height + self.scroll_margin + 1)

        if scroll_offset != self.scroll_offset:
            self.must_repaint = True
        changed = selected_task != self.selected_task or scroll_offset != self.scroll_offset
        self.selected_task = selected_task
        self.scroll_offset = scroll_offset
        return changed

    def run(self):
        self.build_screen()
        curses.doupdate()
        self.update_tasks()

        self.scr.timeout(10000)
        then = datetime.datetime.now()

        while not self.quit:
            source_changed = False

            # detect changes of the sources
            for source in self.sources:
                if source.refresh():
                    source.parse()
                    source_changed = True

            # detect the passing of midnight
            now = datetime.datetime.now()
            past_midnight = then.day != now.day
            then = now

            if source_changed or past_midnight:
                self.must_repaint = True
                self.update_tasks()

            if self.recalculate_scrolling() or self.must_repaint:
                self.paint_tasks(self.must_repaint if isinstance(self.must_repaint, list) else None)
                self.must_repaint = False

            curses.doupdate()

            key = Key.read(self.scr)

            if key == Key.RESIZE:
                self.do_refresh_screen()
            elif not key.special and str(key) in string.digits:
                self.do_jump_to(str(key))
            elif str(key) in self.key_mapping and self.key_mapping[str(key)] in self.functions:
                # clear status bar
                self.paint_status_bar()

                self.functions[self.key_mapping[str(key)]]()

        self.scr.clear()
        self.scr.refresh()

    def read_jump_to(self, start):
        text = tr('Jump to:') + ' '
        self.paint_status_bar(text)
        curses.doupdate()
        number = self.read_text(self.dim[0]-2, len(text), text=start)
        self.paint_status_bar()

        if number is not None and number.isdigit():
            number = int(number)
            return number
        return None

    def read_search_input(self):
        backup_search = self.search.text
        new_search = self.search.text
        curpos = len(new_search)

        def update(text):
            self.search.text = text
            self.search.parse()
            self.apply_search()
            self.paint_tasks()

        new_search = self.read_text(0, 0, text=self.search.text, callback=update)

        if new_search is None:
            new_search = backup_search

        if new_search.strip() != self.search.text:
            self.search.text = new_search.strip()
            self.search.parse()
        self.apply_search()
        self.paint_search_bar()
        self.paint_tasks()

    def read_text(self, y, x, cols=None, text='', callback=None, select_source=False, box=False, title=None):
        if cols is None:
            cols = curses.COLS-x-1
        if box:
            editor = curses.newwin(3, cols, y, x)
        else:
            editor = curses.newwin(1, cols, y, x)
        editor.bkgdset(' ', self.color(common.SETTING_COL_NORMAL))
        editor.erase()
        editor.noutrefresh()

        if box:
            editor.border()
            if title is not None and cols - 8 > len(title):
                editor.addstr(0, 2, f'┤{title}├')
            y = 1
            x = 1
            max_width = cols-3
        else:
            y = 0
            x = 0
            max_width = cols-1
        editor.noutrefresh()

        scroll = 0
        margin = 5
        curpos = len(text)
        source = self.sources[0]

        show_select_source = select_source and box and len(self.sources) > 1

        shortcuts = ['cancel', 'submit-input']
        if show_select_source:
            shortcuts.append('select-file')
        self.paint_shortcut_bar(shortcuts, self.editor_key_mapping)

        def refresh_source_display():
            if not show_select_source:
                return
            right = cols - 3 - len(source.displayname)
            editor.addstr(2, 1, "─"*(cols-2))
            editor.addstr(2, right, f"┤{source.displayname}├")
            editor.noutrefresh()

        refresh_source_display()

        try:
            curses.curs_set(True)
        except:
            pass

        while True:
            unchanged = text
            editor.addstr(y, x, " "*(cols-2))
            editor.noutrefresh()
            if curpos >= scroll + max_width - 2 - margin:
                scroll = min(curpos-margin, curpos - max_width + margin + 2)
            elif curpos <= scroll + margin:
                scroll = max(0, curpos - margin)
            if scroll > 0:
                editor.addstr(y, x, self.overflow_marker[0], self.color(common.SETTING_COL_OVERFLOW))
            if len(text)-scroll >= max_width-1:
                editor.addstr(y, x+max_width-len(self.overflow_marker[1]),
                              self.overflow_marker[1], self.color(common.SETTING_COL_OVERFLOW))
            editor.addstr(y, x + 1, text[scroll:scroll+max_width-2])
            editor.move(y, x + 1 + curpos - scroll)
            editor.refresh()

            key = Key.read(self.scr)

            if str(key) in self.editor_key_mapping:
                fnc = self.editor_key_mapping[str(key)]

                if fnc == 'cancel':
                    text = None
                    break
                elif fnc == 'del-left' and curpos > 0:
                    text = text[:curpos-1] + text[curpos:]
                    curpos = max(0, curpos-1)
                elif fnc == 'del-right':
                    text = text[:curpos] + text[curpos+1:]
                elif fnc == 'del-to-bol':
                    text = text[curpos:]
                    curpos = 0
                elif fnc == 'go-left':
                    curpos = max(0, curpos-1)
                elif fnc == 'go-right':
                    curpos = min(curpos+1, len(text))
                elif fnc == 'go-bol':
                    curpos = 0
                elif fnc == 'go-eol':
                    curpos = len(text)
                elif show_select_source and fnc == 'select-file':
                    curses.curs_set(0)
                    sourcename = self.select_one([s.displayname for s in self.sources], tr("Select source"))
                    curses.curs_set(1)
                    if sourcename is not None:
                        candidates = [s for s in self.sources if s.displayname == sourcename]
                        if len(candidates) == 1:
                            source = candidates[0]
                    self.scr.touchwin()
                    self.paint_shortcut_bar(shortcuts, self.editor_key_mapping)
                    self.scr.noutrefresh()
                    editor.touchwin()
                    editor.noutrefresh()
                    refresh_source_display()
                elif fnc == 'submit-input':
                    break
            elif not key.special:
                text = text[:curpos] + str(key) + text[curpos:]
                curpos += len(key)

            if unchanged != text and callback is not None:
                callback(text)

        try:
            curses.curs_set(0)
        except:
            pass

        del editor
        self.scr.touchwin()
        self.paint_shortcut_bar()
        self.scr.noutrefresh()

        if select_source:
            return text, source
        return text

    def select_one(self, items, title=None, lblfnc=None, autoselect=True):
        if len(items) == 0:
            return None
        if len(items) == 1 and autoselect:
            return items[0]

        if lblfnc is None:
            lblfnc = lambda x: x

        self.paint_shortcut_bar(['cancel', 'select-item', 'next-item', 'prev-item', 'jump-to'])

        options = items[:]
        options.sort(key=lblfnc)
        
        y = curses.LINES//2 - 1
        w_offset = 3 + len(self.selection_indicator) + \
                   len(str(len(options)))+1 if self.show_numbers else 0
        w = min(curses.COLS-6, w_offset + max([len(lblfnc(o)) for o in options]))
        h = min(curses.LINES-6, len(options))
        frame = curses.newwin(h+2, w+2,
                              curses.LINES//2 - h//2 - 1,
                              curses.COLS//2 - w//2 - 1)
        frame.bkgdset(' ', self.color(common.SETTING_COL_NORMAL))
        panel = curses.panel.new_panel(frame)

        scroll = 0
        margin = 2
        selection = 0

        while True:
            if selection < scroll + margin:
                scroll = max(0, selection - margin)
            elif selection >= scroll + h - margin:
                scroll = min(max(0, len(options) - h), selection - h + margin + 1)
            frame.erase()
            frame.border()
            if title is not None and w > len(title) + 6:
                frame.addstr(0, 2, f'┤{title}├')

            for nr, option in enumerate(options):
                if nr < scroll:
                    continue
                if nr-scroll >= h:
                    continue
                x = 1
                y = 1 + nr - scroll
                line = ''

                attrs = self.color(common.SETTING_COL_NORMAL, nr == selection)

                if len(self.selection_indicator) > 0:
                    if selection == nr:
                        indicator = f' {self.selection_indicator}'
                    else:
                        indicator = ' '*(len(self.selection_indicator)+1)
                    frame.addstr(y, x, indicator, attrs)
                    line += indicator
                    x += len(indicator)

                if self.show_numbers:
                    nrwidth = len(str(len(options)))+1
                    nrlbl = f' {nr+1:> {nrwidth}} '
                    frame.addstr(y, x, nrlbl, attrs)
                    line += nrlbl
                    x += len(nrlbl)

                frame.addstr(y, x, option[:max(0, w-len(line))], attrs)
            frame.refresh()

            key = Key.read(self.scr)

            if str(key) in self.key_mapping:
                fnc = self.key_mapping[str(key)]

                if fnc == 'next-item':
                    selection = min(selection + 1, len(options)-1)
                elif fnc == 'prev-item':
                    selection = max(selection - 1, 0)
                elif fnc == 'jump-to':
                    pos = self.read_jump_to('')
                    if pos is not None:
                        selection = min(max(0, pos-1), len(options)-1)
                elif fnc == 'select-item':
                    break
                elif fnc in ['quit', 'cancel']:
                    selection = None
                    break
            elif str(key) in '123456789':
                pos = self.read_jump_to(str(key))
                if pos is not None:
                    selection = min(max(0, pos-1), len(options)-1)

        panel.hide()
        curses.panel.update_panels()
        del panel
        del frame

        if selection is None:
            return None
        return options[selection]

    def do_quit(self):
        self.quit = True

    def do_next_task(self, skip=1):
        selected_task = min(self.selected_task + skip, len(self.filtered_tasks)-1)
        if selected_task != self.selected_task:
            self.must_repaint = [selected_task, self.selected_task]
        self.selected_task = selected_task

    def do_prev_task(self, skip=1):
        selected_task = max(self.selected_task - skip, 0)
        if selected_task != self.selected_task:
            self.must_repaint = [selected_task, self.selected_task]
        self.selected_task = selected_task

    def do_half_page_up(self):
        self.do_prev_task(skip=(self.dim[0] // 2)-1)
        if self.must_repaint:
            self.must_repaint = True

    def do_half_page_down(self):
        self.do_next_task(skip=(self.dim[0] // 2)-1)
        if self.must_repaint:
            self.must_repaint = True

    def do_go_last_task(self):
        self.do_next_task(skip=len(self.filtered_tasks))
        if self.must_repaint:
            self.must_repaint = True

    def do_go_first_task(self):
        self.do_prev_task(skip=len(self.filtered_tasks))
        if self.must_repaint:
            self.must_repaint = True

    def do_jump_to(self, init=''):
        pos = self.read_jump_to(init)
        if pos is not None:
            selected_task = min(max(0, pos-1), len(self.filtered_tasks))
            self.must_repaint = selected_task != self.selected_task
            self.selected_task = selected_task

    def do_start_search(self):
        self.read_search_input()

    def do_search_context(self):
        if len(self.filtered_tasks) == 0 or self.selected_task >= len(self.filtered_tasks):
            return
        task, source = self.filtered_tasks[self.selected_task]
        contexts = [context for context in task.contexts if context not in self.search.contexts]

        if len(contexts) == 0:
            return

        contexts.sort()
        context = self.select_one(contexts, tr("Select Context"))
        self.scr.touchwin()
        self.paint_shortcut_bar()
        self.scr.noutrefresh()
        curses.doupdate()

        if context is not None:
            context = '@' + context
            if len(self.search.text) > 0:
                context = " " + context
            self.search.text += context
            self.search.parse()
            self.apply_search()
            self.paint_search_bar()

    def do_search_project(self):
        if len(self.filtered_tasks) == 0 or self.selected_task >= len(self.filtered_tasks):
            return
        task, source = self.filtered_tasks[self.selected_task]
        projects = [project for project in task.projects if project not in self.search.projects]

        if len(projects) == 0:
            return

        projects.sort()
        project = self.select_one(projects, tr("Select Project"))
        self.scr.touchwin()
        self.paint_shortcut_bar()
        self.scr.noutrefresh()

        if project is not None:
            project = '+' + project
            if len(self.search.text) > 0:
                project = " " + project
            self.search.text += project
            self.search.parse()
            self.apply_search()
            self.paint_search_bar()

    def do_toggle_done(self):
        if len(self.filtered_tasks) == 0 or self.selected_task >= len(self.filtered_tasks):
            return
        task, source = self.filtered_tasks[self.selected_task]
        task = utils.ensure_up_to_date(task)
        if task is not None:
            utils.toggle_done(task, self.clear_contexts)
            source.save(safe=self.safe_save)
        else:
            self.paint_status_bar(tr("Not changed: task was modified in the background"),
                                  self.color(common.SETTING_COL_ERROR))
        self.update_tasks()

    def do_toggle_hidden(self):
        if len(self.filtered_tasks) == 0 or self.selected_task >= len(self.filtered_tasks):
            return
        task, source = self.filtered_tasks[self.selected_task]
        task = utils.ensure_up_to_date(task)
        if task is not None:
            utils.toggle_hidden(task)
            source.save(safe=self.safe_save)
        else:
            self.paint_status_bar(tr("Not changed: task was modified in the background"),
                                  self.color(common.SETTING_COL_ERROR))
        self.update_tasks()

    def do_create_task(self):
        source = self.sources[0]
        y = curses.LINES//2 - 1

        today = ''
        if self.add_creation_date:
            today = datetime.datetime.now().strftime(Task.DATE_FMT) + ' '
        if self.create_from_search:
            today += utils.create_from_search(self.search)
        if self.selected_template is not None:
            if not today.endswith(' '):
                today += ' '
            today += self.selected_template
        text = None
        source = None
        text, source = self.read_text(y, 4, self.dim[1]-8, today,
                                      select_source=True, box=True, title=tr("New Task"))
        if text is not None:
            if self.auto_id and not any([word.startswith('id:') for word in text.split(' ')]):
                text += ' id:#'
            text = utils.auto_task_id(self.sources, text)
            text = utils.dehumanize_dates(text)

        if text is not None and len(text.strip()) > 0:
            task = Task(text)
            if source.refresh():
                source.parse()
            source.tasks.append(task)
            source.save(safe=self.safe_save)
            source.parse()
            source.update_contexts_and_projects()
            self.search.update_sources(self.sources)
            self.update_tasks()
        self.scr.touchwin()
        self.scr.noutrefresh()

    def do_edit_task(self):
        if len(self.filtered_tasks) == 0 or self.selected_task >= len(self.filtered_tasks):
            return
        y = self.dim[0]//2 - 1

        task, source = self.filtered_tasks[self.selected_task]
        if source.refresh():
            source.parse()
            self.update_tasks()
            raw = str(task).strip()
            these = [other for other in source.tasks if str(other).strip() == raw]
            if len(these) > 0:
                task = these[0]
            else:
                self.paint_status_bar(tr("Cannot edit: task was modified in the background"),
                                      self.color(common.SETTING_COL_ERROR))
                return
        text = self.read_text(y, 4, self.dim[1]-8, str(task).strip(),
                              box=True, title=tr("Edit task"))
        if text is not None:
            text = utils.auto_task_id(self.sources, text)
            text = utils.dehumanize_dates(text)

        if text is not None and text.strip() != str(task).strip():
            task = utils.ensure_up_to_date(task)
            if task is not None:
                task.parse(text)
                source.save(safe=self.safe_save)
                source.update_contexts_and_projects()
                self.search.update_sources(self.sources)
            else:
                self.paint_status_bar(tr("Not changed: task was modified in the background"),
                                      self.color(common.SETTING_COL_ERROR))
            self.update_tasks()

    def do_load_template(self):
        templates = utils.parse_templates()
        if len(templates) == 0:
            return
        names = ['None'] + [name for name in sorted(templates.keys())]
        name = self.select_one(names, tr('Load Template'), autoselect=False)
        if names is not None and name in templates:
            self.selected_template = templates[name]
        elif name == 'None':
            self.selected_template = None

    def do_save_template(self):
        y = curses.LINES//2 - 1
        frame = curses.newwin(3, curses.COLS-6, y, 3)
        frame.bkgdset(' ', self.color(common.SETTING_COL_NORMAL))
        frame.erase()
        frame.border()
        label = tr('Save Template')
        if curses.COLS-8 > len(label):
            frame.addstr(0, 2, f'┤{label}├')
        frame.refresh()

        text = self.read_text(y+1, 4, curses.COLS-8, '')

        if text is not None and len(text.strip()) > 0:
            templates = utils.parse_templates()

            if not (len(self.filtered_tasks) == 0 or self.selected_task >= len(self.filtered_tasks)):
                task, source = self.filtered_tasks[self.selected_task]
                raw = str(task).strip().split(' ')
                if len(raw) > 1:
                    raw = ' '.join(raw[1:]) #remove creation date
                    templates[text] = raw
                    utils.save_templates(templates)

        frame.erase()
        del frame

    def do_load_search(self):
        searches = utils.parse_searches()
        if len(searches) == 0:
            return
        names = [name for name in sorted(searches.keys())]
        name = self.select_one(names, tr('Load Search'), autoselect=False)
        self.scr.touchwin()
        self.paint_shortcut_bar()
        self.scr.noutrefresh()
        if names is not None and name in searches:
            self.search.text = searches[name]
            self.paint_search_bar()
            self.search.parse()
            self.apply_search()

    def do_save_search(self):
        y = curses.LINES//2 - 1
        frame = curses.newwin(3, curses.COLS-6, y, 3)
        frame.bkgdset(' ', self.color(common.SETTING_COL_NORMAL))
        frame.erase()
        frame.border()
        label = tr('Save Search')
        if curses.COLS-8 > len(label):
            frame.addstr(0, 2, f'┤{label}├')
        frame.refresh()

        text = self.read_text(y+1, 4, curses.COLS-8, '')

        if text is not None and len(text.strip()) > 0:
            searches = utils.parse_searches()
            searches[text] = self.search.text
            utils.save_searches(searches)

        frame.erase()
        del frame

    def do_reload_tasks(self):
        for source in self.sources:
            source.parse()
            source.update_contexts_and_projects()
        self.search.update_sources(self.sources)
        self.update_tasks()

    def do_toggle_tracking(self):
        if len(self.filtered_tasks) == 0 or self.selected_task >= len(self.filtered_tasks):
            return
        task, source = self.filtered_tasks[self.selected_task]
        task = utils.ensure_up_to_date(task)
        if task is None:
            self.paint_status_bar(tr("Not tracked: task was modified in the background"),
                                  self.color(common.SETTING_COL_ERROR))
            self.update_tasks()
            return
        
        if utils.toggle_tracking(task):
            source.save(safe=self.safe_save)

        self.update_tasks()

    def do_refresh_screen(self):
        curses.update_lines_cols()
        self.dim = self.scr.getmaxyx()
        self.must_repaint = True
        self.build_screen()
        self.paint_search_bar()

    def do_open_url(self):
        if len(self.filtered_tasks) == 0 or self.selected_task >= len(self.filtered_tasks):
            return

        task, _ = self.filtered_tasks[self.selected_task]
        urls = []

        if task.description is None:
            return

        for match in common.URL_RE.finditer(task.description):
            if match.group(1) not in self.protos:
                continue
            urls.append(match.group(0))

        copy = urls[:]
        urls = []
        for url in copy:
            if url.startswith('<') and url.endswith('>'):
                url = url[1:-1]
            if len(url) == 0:
                continue
            urls.append(url)

        url = self.select_one(urls, tr("Select URL"))
        self.scr.touchwin()
        self.paint_shortcut_bar()
        self.scr.noutrefresh()
        if url is not None:
            webbrowser.open(url)

    def do_delegate(self):
        if len(self.filtered_tasks) == 0 or self.selected_task >= len(self.filtered_tasks):
            return

        task, source = self.filtered_tasks[self.selected_task]

        if len(self.delegation_marker) == 0:
            # no delegation marker
            return

        if task.description is None:
            task.description = ''

        if self.delegation_marker in task.description.split(' '):
            # already delegated
            self.do_delegation_action(task)
            return
        
        task = utils.ensure_up_to_date(task)
        if task is not None:
            task.description += ' ' + self.delegation_marker.strip()
            source.save(safe=self.safe_save)
            self.do_delegation_action(task)
        else:
            self.paint_status_bar(tr("Not delegated: task was modified in the background"),
                                  self.color(common.SETTING_COL_ERROR))
        self.update_tasks()

    def do_delegation_action(self, task):
        utils.execute_delegate_action(task, self.delegate_to, self.delegation_marker, self.delegation_action)

    def do_show_help(self):
        def collect_name_fnc(fncs, mapping):
            for name, fnc in sorted([(v, k) for k, v in self.short_name.items() if k in fncs]):
                for key in [k for k, v in mapping.items() if v == fnc]:
                    yield name, key

        nav_fncs = ['next-item', 'prev-item', 'half-page-up', 'half-page-down',
                    'first-item', 'last-item', 'jump-to']
        edt_fncs = ['toggle-hidden', 'toggle-done', 'edit-task', 'create-task',
                    'toggle-tracking', 'delegate', 'save-template', 'load-template']
        search_fncs = ['search', 'load-search', 'save-search', 'search-context', 'search-project']
        meta_fncs = ['show-help', 'open-manual', 'quit', 'cancel', 'refresh-screen',
                     'reload-tasks']
        other_fncs = ['open-url']

        lines = [(tr('TASK LIST'), ''), ('', ''), (tr('Program'), '')]
        lines += [(name, key) for name, key in collect_name_fnc(meta_fncs, self.key_mapping)]
        lines += [('', ''), (tr('Navigation'), '')]
        lines += [(name, key) for name, key in collect_name_fnc(nav_fncs, self.key_mapping)]
        lines += [('', ''), (tr('Search'), '')]
        lines += [(name, key) for name, key in collect_name_fnc(search_fncs, self.key_mapping)]
        lines += [('', ''), (tr('Change Tasks'), '')]
        lines += [(name, key) for name, key in collect_name_fnc(edt_fncs, self.key_mapping)]
        lines += [('', ''), (tr('Other'), '')]
        lines += [(name, key) for name, key in collect_name_fnc(other_fncs, self.key_mapping)]

        edt_nav_fncs = ['go-left', 'go-right', 'go-bol', 'go-eol']
        edt_edt_fncs = ['del-left', 'del-right', 'del-to-bol']
        edt_meta_fncs = ['cancel', 'submit-input']
        lines += [('', ''), ('', ''), (tr('TASK EDITING'), ''), ('', ''), (tr('General'), '')]
        lines += [(name, key) for name, key in collect_name_fnc(edt_meta_fncs, self.editor_key_mapping)]
        lines += [('', ''), (tr('Navigation'), '')]
        lines += [(name, key) for name, key in collect_name_fnc(edt_nav_fncs, self.editor_key_mapping)]
        lines += [('', ''), (tr('Deletion'), '')]
        lines += [(name, key) for name, key in collect_name_fnc(edt_edt_fncs, self.editor_key_mapping)]

        maxnamelen = max([len(n) for n, _ in lines])
        scroll = 0
        doupdate = 2

        while True:
            if doupdate > 0:
                if doupdate > 1:
                    curses.update_lines_cols()
                    frame = curses.newwin(curses.LINES-1, curses.COLS, 0, 0)
                    frame.clear()
                    frame.bkgd(' ', self.color(common.SETTING_COL_NORMAL))
                    self.paint_shortcut_bar(['quit',])
                else:
                    frame.clear()
                maxscroll = max(0, len(lines) - frame.getmaxyx()[0] + 3)
                scroll = min(scroll, maxscroll)
                for idx, line in enumerate(lines):
                    if idx < scroll:
                        continue
                    if idx >= scroll + frame.getmaxyx()[0] - 2:
                        continue
                    attrs = common.SETTING_COL_NORMAL
                    if line[1] == '':
                        attrs = common.SETTING_COL_CONTEXT
                    frame.addstr(idx+1-scroll, 1, line[0] + " "*(maxnamelen-len(line[0])+3) + line[1], self.color(attrs))
                    frame.clrtoeol()
                frame.clrtobot()
                frame.border()
                frame.noutrefresh()
                curses.doupdate()
                doupdate = 0

            key = Key.read(self.scr)

            if key == Key.RESIZE:
                doupdate = 2
                continue

            if str(key) in self.key_mapping:
                fnc = self.key_mapping[str(key)]

                if fnc in ['quit', 'cancel']:
                    break
                elif fnc == 'refresh-screen':
                    doupdate = 2
                elif fnc == 'next-item':
                    scroll = min(scroll + 1, maxscroll)
                    doupdate = 1
                elif fnc == 'prev-item':
                    scroll = max(0, scroll - 1)
                    doupdate = 1
                elif fnc == 'first-item':
                    scroll = 0
                    doupdate = 1
                elif fnc == 'last-item':
                    scroll = maxscroll
                    doupdate = 1
                elif fnc == 'half-page-up':
                    scroll = max(0, scroll - frame.getmaxyx()[0]//2)
                    doupdate = 1
                elif fnc == 'half-page-down':
                    scroll = min(maxscroll, scroll + frame.getmaxyx()[0]//2)
                    doupdate = 1

        del frame
        self.do_refresh_screen()


def run_cursesui(args):
    success = 0
    sources = utils.open_sources(args)
    
    if len(sources) == 0:
        success = -1
        print(tr("To start pter you must provide at least one todo.txt file. See --help for more information."),
              file=sys.stderr)
    else:
        search = Searcher('', False)
        search.update_sources(sources)

        window = Window(sources, configuration.get_config(args))
        exception = None

        try:
            window.run()
        except Exception as exc:
            exception = exc
        
        curses.endwin()

        if exception is not None:
            raise exception
            success = -3

    return success




import datetime
import string

from pytodotxt import Task


class Searcher:
    def __init__(self, text=None, casesensitive=True, default_threshold=None, hide_sequential=True):
        self.words = set()
        self.not_words = set()
        self.projects = set()
        self.not_projects = set()
        self.contexts = set()
        self.not_contexts = set()
        self.ids = set()
        self.not_ids = set()
        self.filenames = set()
        self.not_filenames = set()
        self.after = set()
        self.refs = set()
        self.done = None
        self.show_hidden = False
        self.priority = None
        self.default_threshold = default_threshold
        self.hide_sequential = hide_sequential
        if default_threshold is None or len(default_threshold) == 0:
            self.default_threshold = 'today'
        self.threshold = self.default_threshold
        self.due = None
        self.created = None
        self.completed = None
        self.casesensitive = casesensitive
        self.sources = None
        self.task_by_id = None

        self.text = text

        if text is not None:
            self.parse()

    def reset(self):
        self.words = set()
        self.not_words = set()
        self.projects = set()
        self.not_projects = set()
        self.contexts = set()
        self.not_contexts = set()
        self.ids = set()
        self.not_ids = set()
        self.filenames = set()
        self.not_filenames = set()
        self.after = set()
        self.refs = set()
        self.done = None
        self.show_hidden = False
        self.priority = None
        self.threshold = get_relative_date(self.default_threshold, Task.DATE_FMT) or self.default_threshold
        self.due = None
        self.created = None
        self.completed = None

    def update_sources(self, sources):
        self.sources = sources
        self.task_by_id = {}
        self.parents = {}
        with_sequence = set()
        for source in self.sources:
            for task in source.tasks:
                if len(task.attr_id) == 0:
                    continue
                if len(task.attr_after) > 0:
                    with_sequence.add(task)
                for taskid in task.attr_id:
                    if taskid not in self.task_by_id:
                        self.task_by_id[taskid] = set()
                    self.task_by_id[taskid].add(task)

        # for each task obtain a set of their direct and indirect parents
        for task in with_sequence:
            parents = self._parse_ids('after', task)
            queue = parents.copy()
            while len(queue) > 0:
                otherid = queue.pop()
                if otherid in self.parents:
                    parents |= self.parents[otherid]
                    break
                parents.add(otherid)

                if otherid not in self.task_by_id:
                    continue
                for other in self.task_by_id[otherid]:
                    queue |= self._parse_ids('after', other).difference(parents)
            if len(parents) > 0:
                for taskid in task.attr_id:
                    if taskid not in self.parents:
                        self.parents[taskid] = set()
                    self.parents[taskid] |= parents

    def _parse_ids(self, keyword, task):
        these = set()
        for value in task.attributes.get(keyword, []):
            these |= {that for that in value.split(',')}
        if self.casesensitive:
            return {that.lower() for that in these}
        else:
            return these

    def parse(self):
        self.reset()
        text = self.text
        if not self.casesensitive:
            text = text.lower()

        for part in text.split(' '):
            do_not = False

            if part.startswith('sort:'):
                continue

            if part.startswith('not:'):
                do_not = True
                part = part[4:]
            elif part.startswith('-'):
                do_not = True
                part = part[1:]

            if len(part) == 0:
                continue

            if part.startswith('@') and len(part) > 1:
                if do_not:
                    self.not_contexts.add(part[1:])
                else:
                    self.contexts.add(part[1:])
            
            elif part.startswith('+') and len(part) > 1:
                if do_not:
                    self.not_projects.add(part[1:])
                else:
                    self.projects.add(part[1:])
            
            elif part.startswith('done:'):
                _, value = part.split(':', 1)
                if len(value) == 0:
                    self.done = None
                else:
                    self.done = value.lower().startswith('y')
            
            elif part.startswith('hidden:') or part.startswith('h:'):
                _, value = part.split(':', 1)
                self.show_hidden = value.lower().startswith('y') or value == '1'
            
            elif part.startswith('pri:'):
                _, value = part.split(':', 1)
                self.priority = (value, value)
            
            elif part.startswith('moreimportant:') or part.startswith('mi:'):
                _, value = part.split(':', 1)
                if self.priority is None:
                    self.priority = ['ZZZZ', ' ']
                self.priority[0] = value.upper()
            
            elif part.startswith('lessimportant:') or part.startswith('li:'):
                _, value = part.split(':', 1)
                if self.priority is None:
                    self.priority = ['ZZZZ', ' ']
                self.priority[1] = value.upper()
            
            elif part.startswith('due:'):
                _, value = part.split(':', 1)
                if value.lower().startswith('y'):
                    self.due = ['0000-00-00', '9999-99-99']
                else:
                    self.due = [value, value]
            
            elif part.startswith('duebefore:') or part.startswith('db:'):
                _, value = part.split(':', 1)
                if self.due is None:
                    self.due = ['0000-00-00', '9999-99-99']
                self.due[1] = value
            
            elif part.startswith('dueafter:') or part.startswith('da:'):
                _, value = part.split(':', 1)
                if self.due is None:
                    self.due = ['0000-00-00', '9999-99-99']
                self.due[0] = value

            elif part.startswith('created:'):
                _, value = part.split(':', 1)
                self.created = [value, value]

            elif part.startswith('createdbefore:') or part.startswith('crb:'):
                _, value = part.split(':', 1)
                if self.created is None:
                    self.created = ['0000-00-00', '9999-99-99']
                self.created[1] = value

            elif part.startswith('createdafter:') or part.startswith('cra:'):
                _, value = part.split(':', 1)
                if self.created is None:
                    self.created = ['0000-00-00', '9999-99-99']
                self.created[0] = value

            elif part.startswith('completed:'):
                _, value = part.split(':', 1)
                self.completed = [value, value]

            elif part.startswith('completedbefore:') or part.startswith('cob:'):
                _, value = part.split(':', 1)
                if self.completed is None:
                    self.completed = ['0000-00-00', '9999-99-99']
                self.completed[1] = value

            elif part.startswith('completedafter:') or part.startswith('coa:'):
                _, value = part.split(':', 1)
                if self.completed is None:
                    self.completed = ['0000-00-00', '9999-99-99']
                self.completed[0] = value

            elif part.startswith('t:') or part.startswith('tickler:') or part.startswith('threshold:'):
                _, value = part.split(':', 1)
                if len(value) == 0:
                    self.threshold = None
                else:
                    self.threshold = get_relative_date(value, Task.DATE_FMT) or value

            elif part.startswith('id:'):
                _, value = part.split(':', 1)
                if len(value) > 0:
                    value = {that for that in value.split(',')}

                    if do_not:
                        self.not_ids |= value
                    else:
                        self.ids |= value
                else:
                    if do_not:
                        self.not_words.add('id:')
                    else:
                        self.words.add('id:')

            elif part.startswith('after:') and self.after is not None:
                _, value = part.split(':', 1)
                if len(value) == 0:
                    self.after = None
                else:
                    values = set(value.split(','))
                    if do_not:
                        pass
                    else:
                        self.after |= values

            elif part.startswith('ref:') and self.refs is not None:
                _, value = part.split(':', 1)
                if len(value) > 0:
                    values = set(value.split(','))
                    if do_not:
                        pass
                    else:
                        self.refs |= values

            elif part.startswith('file:'):
                _, value = part.split(':', 1)
                if len(value) > 0:
                    if do_not:
                        self.not_filenames.add(value)
                    else:
                        self.filenames.add(value)
            
            else:
                if do_not:
                    self.not_words.add(part)
                else:
                    self.words.add(part)

    def match(self, task):
        attrs = dict([(k if self.casesensitive else k.lower(), v)
                      for k, v in task.attributes.items()])
        return all([self.match_words(task),
                    self.match_contexts(task),
                    self.match_projects(task),
                    self.match_done(task),
                    self.match_hidden(attrs),
                    self.match_priority(task),
                    self.match_ids(task),
                    self.match_filenames(task),
                    self.match_refs(task),
                    self.match_after(task, attrs),
                    self.match_due(attrs),
                    self.match_created(task),
                    self.match_completed(task),
                    self.match_threshold(attrs)])

    def match_words(self, task):
        if len(self.words) == 0 and len(self.not_words) == 0:
            return True

        description = task.description
        if description is None:
            description = ''

        if not self.casesensitive:
            description = description.lower()

        return all([word in description for word in self.words]) \
                and not any([word in description for word in self.not_words])

    def match_contexts(self, task):
        if len(self.contexts) == 0 and len(self.not_contexts) == 0:
            return True

        contexts = task.contexts
        if not self.casesensitive:
            contexts = [context.lower() for context in contexts]

        return all([context in contexts for context in self.contexts]) \
                and not any([context in contexts for context in self.not_contexts])

    def match_projects(self, task):
        if len(self.projects) == 0 and len(self.not_projects) == 0:
            return True

        projects = task.projects
        if not self.casesensitive:
            projects = [project.lower() for project in projects]

        return all([project in projects for project in self.projects]) \
                and not any([project in projects for project in self.not_projects])

    def match_hidden(self, attrs):
        return 'h' not in attrs or (attrs['h'][0] == '1') == self.show_hidden

    def match_done(self, task):
        return self.done is None or task.is_completed == self.done

    def match_priority(self, task):
        if self.priority is None:
            return True

        pri = 'ZZZ'
        if task.priority is not None:
            pri = task.priority
        if not self.casesensitive:
            pri = pri.lower()

        return (self.priority[0] == self.priority[1] and pri == self.priority[0]) or \
               self.priority[0] > pri > self.priority[1]

    def match_threshold(self, attrs):
        if self.threshold is None:
            return True
        return 't' not in attrs or attrs['t'][0] <= self.threshold

    def match_due(self, attrs):
        if self.due is None:
            return True

        if 'due' not in attrs:
            return False

        due = [get_relative_date(self.due[0], Task.DATE_FMT) or self.due[0],
               get_relative_date(self.due[1], Task.DATE_FMT) or self.due[1]]

        if due[0] == due[1]:
            return attrs['due'][0] == due[0]

        return due[0] < attrs['due'][0] < due[1]

    def match_created(self, task):
        if self.created is None:
            return True

        if task.creation_date is None:
            return False

        created = [get_relative_date(self.created[0], Task.DATE_FMT) or self.created[0],
                   get_relative_date(self.created[1], Task.DATE_FMT) or self.created[1]]
        task_created = task.creation_date.strftime(Task.DATE_FMT)

        if created[0] == created[1]:
            return created[0] == task_created

        return created[0] < task_created < created[1]

    def match_completed(self, task):
        if self.completed is None:
            return True

        if task.completion_date is None:
            return False

        completed = [get_relative_date(self.completed[0], Task.DATE_FMT) or self.completed[0],
                     get_relative_date(self.completed[1], Task.DATE_FMT) or self.completed[1]]
        task_completed = task.completion_date.strftime(Task.DATE_FMT)

        if completed[0] == completed[1]:
            return completed[0] == task_completed

        return completed[0] < task_completed < completed[1]

    def match_ids(self, task):
        if len(self.not_ids) == 0 and len(self.ids) == 0:
            return True

        ids = self._parse_ids('id', task)

        return (len(self.ids) == 0 or len(self.ids.intersection(ids)) > 0) and \
               (len(self.not_ids) == 0 or len(self.not_ids.intersection(ids)) == 0)

    def match_refs(self, task):
        if len(self.refs) == 0:
            return True

        ids = self._parse_ids('after', task)
        ids |= self._parse_ids('ref', task)

        return len(ids) > 0 and \
               len(self.refs.intersection(ids)) > 0

    def match_filenames(self, task):
        if len(self.filenames) + len(self.not_filenames) == 0:
            return True

        return hasattr(task, 'todotxt') is not None and \
               task.todotxt is not None and \
               all([pattern in str(task.todotxt.filename) for pattern in self.filenames]) and \
               not any([pattern in str(task.todotxt.filename) for pattern in self.not_filenames])

    def match_after(self, task, attrs):
        if self.after is None:
            return True

        if len(self.after) == 0 and \
           len(attrs.get('after', [])) == 0:
            return True

        if self.task_by_id is None:
            return True

        parents = self._parse_ids('after', task)

        if len(self.after) == 0:
            # normal sequence behaviour
            return not self.hide_sequential or \
                   all([parentid not in self.task_by_id or
                        any([other.is_completed for other in self.task_by_id[parentid]]) 
                        for parentid in parents])
        else:
            # collect all indirect parents, too
            for taskid in task.attr_id:
                parents |= self.parents.get(taskid, set())
            return any([parent in parents for parent in self.after])


def get_relative_date(text, fmt=None):
    weekdays = ['mon', 'tue', 'wed', 'thu', 'fri',
                'sat', 'sun']
    days = 0
    weeks = 0
    months = 0
    years = 0
    value = ''
    today = datetime.datetime.now()

    if text.lower() == 'today':
        if fmt is not None:
            return today.strftime(fmt)
        return today
    if text.lower() == 'tomorrow':
        then = today + datetime.timedelta(days=1)
        if fmt is not None:
            return then.strftime(fmt)
        return then
    if text.lower() in ['next-week', 'week']:
        return today + datetime.timedelta(weeks=1)
    if text.lower() in weekdays:
        index = weekdays.index(text.lower())
        days = (index - today.weekday()) % 7
        if days == 0:
            days = 7
        then = today + datetime.timedelta(days=days)
        if fmt is not None:
            return then.strftime(fmt)
        return then

    for char in text.lower():
        if char in string.digits:
            value += char
        elif char in '+-' and len(value) == 0:
            if char == '-':
                value = char
        elif value.isnumeric() or (value.startswith('-') and value[1:].isnumeric()):
            if char == 'd':
                days += int(value)
            elif char == 'w':
                weeks += int(value)
            elif char == 'm':
                months += int(value)
            elif char == 'y':
                years += int(value)
            else:
                # parse error
                return None
            value = ''
        else:
            # parse error
            return None

    if value.isnumeric() or (value.startswith('-') and value[1:].isnumeric()):
        days += int(value)

    month = today.month + months
    if month > 12:
        years += month // 12
        month = month % 12

    year = today.year + years

    if year + years > datetime.MAXYEAR:
        year = datetime.MAXYEAR

    then = datetime.date(year, month, 1) + \
           datetime.timedelta(days=today.day + days - 1,
                              weeks=weeks)
    if fmt is not None:
        then = then.strftime(fmt)
    return then




import unittest
import pathlib
import datetime

from pytodotxt import Task

from pter import utils
from pter import common


class FakeSource:
    def __init__(self, tasks):
        self.tasks = tasks
        self.filename = pathlib.Path('/tmp/test.txt')


class TestSorting(unittest.TestCase):
    def testDueSorting(self):
        tasks = [Task(utils.dehumanize_dates("due in a week due:+1w id:1")),
                 Task(utils.dehumanize_dates("due tomorrow due:1 id:2")),
                 Task(utils.dehumanize_dates("due in a month due:+1m id:3")),
                 Task(utils.dehumanize_dates("(A) no due date id:4"))]
        sortorder = utils.build_sort_order(common.DEFAULT_SORT_ORDER)

        tasks.sort(key=lambda t: utils.sort_fnc(t, sortorder))

        self.assertEqual(tasks[0].attr_id, ['2'])
        self.assertEqual(tasks[1].attr_id, ['1'])
        self.assertEqual(tasks[2].attr_id, ['3'])
        self.assertEqual(tasks[3].attr_id, ['4'])

    def testOverdueSorting(self):
        tasks = [Task(utils.dehumanize_dates("(A) Important id:1")),
                 Task(utils.dehumanize_dates("due tomorrow due:1 id:2")),
                 Task(utils.dehumanize_dates("due yesterday due:-1 id:3")),
                 Task(utils.dehumanize_dates("Not relevant id:4"))]
        sortorder = utils.build_sort_order(common.DEFAULT_SORT_ORDER)

        tasks.sort(key=lambda t: utils.sort_fnc(t, sortorder))

        self.assertEqual(tasks[0].attr_id, ['3'])
        self.assertEqual(tasks[1].attr_id, ['2'])
        self.assertEqual(tasks[2].attr_id, ['1'])
        self.assertEqual(tasks[3].attr_id, ['4'])




"""Common constants for both ncurses and qiui"""
import pathlib
import os
import re

try:
    from xdg import BaseDirectory
except ImportError:
    BaseDirectory = None


PROGRAMNAME = 'pter'
QTPROGRAMNAME = 'qpter'
HERE = pathlib.Path(os.path.abspath(__file__)).parent
HOME = pathlib.Path.home()
CONFIGDIR = HOME / ".config" / PROGRAMNAME
CONFIGFILE = HOME / ".config" / PROGRAMNAME / (PROGRAMNAME + ".conf")
CACHEDIR = HOME / ".cache" / PROGRAMNAME
CACHEFILE = CACHEDIR / (PROGRAMNAME + ".settings")

if BaseDirectory is not None:
    CONFIGDIR = pathlib.Path(BaseDirectory.save_config_path(PROGRAMNAME) or CONFIGDIR)
    CONFIGFILE = CONFIGDIR / (PROGRAMNAME + ".conf")
    CACHEDIR = pathlib.Path(BaseDirectory.save_cache_path(PROGRAMNAME) or CACHEDIR)
    CACHEFILE = CACHEDIR / (PROGRAMNAME + ".settings")

SEARCHES_FILE = CONFIGDIR / "searches.txt"
TEMPLATES_FILE = CONFIGDIR / "templates.txt"

URL_RE = re.compile(r'([A-Za-z][A-Za-z0-9+\-.]*)://([^ ]+)')

DEFAULT_TASK_FORMAT = '{selection: >} {nr: >} {done} {tracking }{due }{(pri) }{description}'
ATTR_TRACKING = 'tracking'
ATTR_T = 't'
ATTR_DUE = 'due'
ATTR_PRI = 'pri'
ATTR_ID = 'id'

DELEGATE_ACTION_NONE = 'none'
DELEGATE_ACTION_MAIL = 'mail-to'
DELEGATE_ACTIONS = (DELEGATE_ACTION_NONE, DELEGATE_ACTION_MAIL)

DEFAULT_SORT_ORDER = 'completed,due_in,priority,linenr'

SETTING_GROUP_GENERAL = 'General'
SETTING_GROUP_SYMBOLS = 'Symbols'
SETTING_GROUP_COLORS = 'Colors'
SETTING_GROUP_HIGHLIGHT = 'Highlight'
SETTING_GROUP_KEYS = 'Keys'
SETTING_GROUP_EDITORKEYS = 'Editor:Keys'
SETTING_GROUP_GUICOLORS = 'GUI:Colors'
SETTING_GROUP_GUIHIGHLIGHT = 'GUI:Highlight'
SETTING_GROUP_GUIKEYS = 'GUI:Keys'
SETTING_GROUP_GUI = 'GUI'
SETTING_GROUP_INCLUDE = 'Include'
SETTING_HUMAN_DATES = 'human-friendly-dates'
SETTING_PROTOCOLS = 'protocols'
SETTING_DELEG_MARKER = 'delegation-marker'
SETTING_DELEG_ACTION = 'delegation-action'
SETTING_DELEG_TO = 'delegation-to'
SETTING_DEFAULT_THRESHOLD = 'default-threshold'
SETTING_ADD_CREATED = 'add-creation-date'
SETTING_SEARCH_CASE_SENSITIVE = 'search-case-sensitive'
SETTING_SAFE_SAVE = 'safe-save'
SETTING_SCROLL_MARGIN = 'scroll-margin'
SETTING_SHOW_NUMBERS = 'show-numbers'
SETTING_USE_COLORS = 'use-colors'
SETTING_TASK_FORMAT = 'task-format'
SETTING_CLEAR_CONTEXT = 'clear-contexts'
SETTING_FONT = 'font'
SETTING_FONTSIZE = 'font-size'
SETTING_SINGLE_INSTANCE = 'single-instance'
SETTING_CREATE_FROM_SEARCH = 'create-from-search'
SETTING_AUTO_ID = 'auto-id'
SETTING_HIDE_SEQUENTIAL = 'hide-sequential'
SETTING_CLICKABLE = 'clickable'
SETTING_DAILY_RELOAD = 'daily-reload'
SETTING_ICON_SELECTION = 'selection'
SETTING_ICON_NOT_DONE = 'not-done'
SETTING_ICON_DONE = 'done'
SETTING_ICON_OVERFLOW_LEFT = 'overflow-left'
SETTING_ICON_OVERFLOW_RIGHT = 'overflow-right'
SETTING_ICON_OVERDUE = 'overdue'
SETTING_ICON_DUE_TODAY = 'due-today'
SETTING_ICON_DUE_TOMORROW = 'due-tomorrow'
SETTING_ICON_TRACKING = 'tracking'
SETTING_COL_NORMAL = 'normal'
SETTING_COL_PRI_A = 'pri-a'
SETTING_COL_PRI_B = 'pri-b'
SETTING_COL_PRI_C = 'pri-c'
SETTING_COL_INACTIVE = 'inactive'
SETTING_COL_CONTEXT = 'context'
SETTING_COL_PROJECT = 'project'
SETTING_COL_ERROR = 'error'
SETTING_COL_HELP_TEXT = 'help'
SETTING_COL_HELP_KEY = 'help-key'
SETTING_COL_OVERFLOW = 'overflow'
SETTING_COL_OVERDUE = 'overdue'
SETTING_COL_DUE_TODAY = 'due-today'
SETTING_COL_DUE_TOMORROW = 'due-tomorrow'
SETTING_COL_TRACKING = 'tracking'
SETTING_COL_URL = 'url'
SETTING_GK_QUIT = 'quit'
SETTING_GK_NEW = 'new'
SETTING_GK_NEW_REF = 'new-related'
SETTING_GK_NEW_AFTER = 'new-subsequent'
SETTING_GK_EDIT = 'edit'
SETTING_GK_OPEN_FILE = 'open-file'
SETTING_GK_TOGGLE_DONE = 'toggle-done'
SETTING_GK_SEARCH = 'search'
SETTING_GK_TOGGLE_TRACKING = 'toggle-tracking'
SETTING_GK_OPEN_MANUAL = 'open-manual'
SETTING_GK_NAMED_SEARCHES = 'named-searches'
SETTING_GK_FOCUS_TASKS = 'focus-tasks'
SETTING_GK_TOGGLE_HIDDEN = 'toggle-hidden'
SETTING_GK_TOGGLE_DARK = 'toggle-dark-mode'
SETTING_GK_DELEGATE = 'delegate'

TF_SELECTION = 'selection'
TF_NUMBER = 'nr'
TF_DESCRIPTION = 'description'
TF_DONE = 'done'
TF_TRACKING = 'tracking'
TF_DUE = 'due'
TF_ALL = 'all'
TF_DUEDAYS = 'duedays'
TF_PRIORITY = 'pri'
TF_CREATED = 'created'
TF_COMPLETED = 'completed'
TF_AGE = 'age'




import curses


class Key:
    SPECIAL = {'<backspace>': '⌫',  # ⌫
               '<del>': '⌦',
               '<ins>': '⎀',
               '<left>': '→',
               '<right>': '←',
               '<up>': '↑',
               '<down>': '↓',
               '<home>': 'Home',
               '<end>': 'End',
               '<escape>': '⎋',
               '<return>': '⏎',  # ↵ ↲
               '<pgup>': 'PgUp',  # ⇞
               '<pgdn>': 'PgDn',  # ⇟
               ' ': '␣',
               '<tab>': '⇥',
               '<f1>': 'F1',
               '<f2>': 'F2',
               '<f3>': 'F3',
               '<f4>': 'F4',
               '<f5>': 'F5',
               '<f6>': 'F6',
               '<f7>': 'F7',
               '<f8>': 'F8',
               '<f9>': 'F9',
               '<f10>': 'F10',
               '<f11>': 'F11',
               '<f12>': 'F12',
               }
    BACKSPACE = '<backspace>'
    DELETE = '<del>'
    LEFT = '<left>'
    RIGHT = '<right>'
    UP = '<up>'
    DOWN = '<down>'
    PGUP = '<pgup>'
    PGDN = '<pgdn>'
    HOME = '<home>'
    END = '<end>'
    RETURN = '<return>'
    ESCAPE = '<escape>'
    SPACE = ' '
    TAB = '<tab>'
    F1 = '<f1>'
    F2 = '<f2>'
    F3 = '<f3>'
    F4 = '<f4>'
    F5 = '<f5>'
    F6 = '<f6>'
    F7 = '<f7>'
    F8 = '<f8>'
    F9 = '<f9>'
    F10 = '<f10>'
    F11 = '<f11>'
    F12 = '<f12>'
    RESIZE = '<resize>'
    TIMEOUT = '<timeout>'

    def __init__(self, value, special=False):
        self.value = value
        self.special = special

    @classmethod
    def read(cls, stdscr):
        try:
            value = stdscr.get_wch()
            if value == -1:
                return Key(Key.TIMEOUT, special=True)
            return Key.parse(value)
        except curses.error as exc:
            if str(exc) == 'no input':
                return Key(Key.TIMEOUT, special=True)
            return Key('C', special=True)
        except KeyboardInterrupt:
            return Key('C', special=True)
        except EOFError:
            return Key('D', special=True)

    @classmethod
    def parse(cls, value):
        if value == curses.KEY_BACKSPACE:
            return Key(Key.BACKSPACE, True)
        elif value == curses.KEY_DC:
            return Key(Key.DELETE, True)
        elif value == curses.KEY_LEFT:
            return Key(Key.LEFT, True)
        elif value == curses.KEY_RIGHT:
            return Key(Key.RIGHT, True)
        elif value == curses.KEY_UP:
            return Key(Key.UP, True)
        elif value == curses.KEY_DOWN:
            return Key(Key.DOWN, True)
        elif value == curses.KEY_END:
            return Key(Key.END, True)
        elif value == curses.KEY_HOME:
            return Key(Key.HOME, True)
        elif value == curses.KEY_NPAGE:
            return Key(Key.PGDN, True)
        elif value == curses.KEY_PPAGE:
            return Key(Key.PGUP, True)
        elif value == curses.KEY_F1:
            return Key(Key.F1, True)
        elif value == curses.KEY_F2:
            return Key(Key.F2, True)
        elif value == curses.KEY_F3:
            return Key(Key.F3, True)
        elif value == curses.KEY_F4:
            return Key(Key.F4, True)
        elif value == curses.KEY_F5:
            return Key(Key.F5, True)
        elif value == curses.KEY_F6:
            return Key(Key.F6, True)
        elif value == curses.KEY_F7:
            return Key(Key.F7, True)
        elif value == curses.KEY_F8:
            return Key(Key.F8, True)
        elif value == curses.KEY_F9:
            return Key(Key.F9, True)
        elif value == curses.KEY_F10:
            return Key(Key.F10, True)
        elif value == curses.KEY_F11:
            return Key(Key.F11, True)
        elif value == curses.KEY_F12:
            return Key(Key.F12, True)
        elif value == curses.KEY_RESIZE:
            return Key(Key.RESIZE, True)
        elif isinstance(value, int):
            # no idea what key that is
            return Key('', True)
        elif isinstance(value, str):
            try:
                ctrlkey = str(curses.unctrl(value), 'ascii')
            except OverflowError:
                # some unicode, you probably want to see it
                return Key(value, False)

            if value in "\n\r":
                return Key(Key.RETURN, special=True)

            if value == ' ':
                return Key(Key.SPACE)

            if value == "\t":
                return Key(Key.TAB)

            if ctrlkey in ['^H', '^?']:
                return Key(Key.BACKSPACE, special=True)

            if ctrlkey == '^[':
                return Key(Key.ESCAPE, True)

            if ctrlkey != value:
                return Key(ctrlkey[1:], True)
            else:
                return Key(value)

    def __len__(self):
        return len(str(self))

    def __eq__(self, other):
        if isinstance(other, Key):
            return self.value == other.value and self.special == other.special
        elif isinstance(other, str):
            return str(self) == other
        elif isinstance(other, bytes):
            return str(self) == str(other, 'ascii')
        raise ValueError("'other' has unexpected type {type(other)}")

    def __str__(self):
        if self.special:
            if self.value.startswith('<'):
                return self.value
            return  '^' + self.value
        return self.value




import pathlib
import configparser
import datetime

from pter import common


DEFAULT_CONFIG = {
        common.SETTING_GROUP_GENERAL: {
            common.SETTING_USE_COLORS: 'yes',
            common.SETTING_SHOW_NUMBERS: 'yes',
            common.SETTING_SCROLL_MARGIN: 5,
            common.SETTING_SAFE_SAVE: 'yes',
            common.SETTING_SEARCH_CASE_SENSITIVE: 'yes',
            common.SETTING_DEFAULT_THRESHOLD: '',
            common.SETTING_HUMAN_DATES: '',
            common.SETTING_DELEG_MARKER: '@delegated',
            common.SETTING_DELEG_ACTION: common.DELEGATE_ACTION_NONE,
            common.SETTING_DELEG_TO: 'to',
            common.SETTING_ADD_CREATED: 'yes',
            common.SETTING_TASK_FORMAT: common.DEFAULT_TASK_FORMAT,
            common.SETTING_CLEAR_CONTEXT: '',
            common.SETTING_PROTOCOLS: 'http,https,mailto,ftp,ftps',
            common.SETTING_CREATE_FROM_SEARCH: 'no',
            common.SETTING_AUTO_ID: 'no',
            common.SETTING_HIDE_SEQUENTIAL: 'yes',
        },
        common.SETTING_GROUP_SYMBOLS: {
            common.SETTING_ICON_SELECTION: '',
            common.SETTING_ICON_NOT_DONE: '[ ]',
            common.SETTING_ICON_DONE: '[x]',
            common.SETTING_ICON_OVERFLOW_LEFT: '←',
            common.SETTING_ICON_OVERFLOW_RIGHT: '→',
            common.SETTING_ICON_OVERDUE: '!!',
            common.SETTING_ICON_DUE_TODAY: '!',
            common.SETTING_ICON_DUE_TOMORROW: '*',
            common.SETTING_ICON_TRACKING: '@',
        },
        common.SETTING_GROUP_COLORS: {
            common.SETTING_COL_NORMAL: '',
            common.SETTING_COL_PRI_A: '',
            common.SETTING_COL_PRI_B: '',
            common.SETTING_COL_PRI_C: '',
            common.SETTING_COL_INACTIVE: '',
            common.SETTING_COL_CONTEXT: '',
            common.SETTING_COL_PROJECT: '',
            common.SETTING_COL_ERROR: '',
            common.SETTING_COL_HELP_TEXT: '',
            common.SETTING_COL_HELP_KEY: '',
            common.SETTING_COL_OVERFLOW: '',
            common.SETTING_COL_OVERDUE: '',
            common.SETTING_COL_DUE_TODAY: '',
            common.SETTING_COL_DUE_TOMORROW: '',
            common.SETTING_COL_TRACKING: '',
        },
        common.SETTING_GROUP_GUICOLORS: {
            'overdue': '#bf360c',
            'due-today': '#f57f17',
            'due-tomorrow': '#827717',
            'done': '#9e9e9e',
            'context': '#1565c0',
            'project': '#00695c',
            'tracking': '#388e3c',
            'pri-a': '#bf360c',
            'pri-b': '#e65100',
            'pri-c': '#ff6f00',
            'url': '#4527a0',
        },
        common.SETTING_GROUP_GUI: {
            common.SETTING_SINGLE_INSTANCE: 'yes',
            common.SETTING_CLICKABLE: 'yes',
            common.SETTING_DAILY_RELOAD: '00:00',
        },
        common.SETTING_GROUP_KEYS: {
        },
        common.SETTING_GROUP_EDITORKEYS: {
        },
        common.SETTING_GROUP_HIGHLIGHT: {
        },
        common.SETTING_GROUP_GUIHIGHLIGHT: {
        },
        common.SETTING_GROUP_GUIKEYS: {
            common.SETTING_GK_QUIT: 'Ctrl+Q',
            common.SETTING_GK_NEW: 'Ctrl+N',
            common.SETTING_GK_NEW_REF: '',
            common.SETTING_GK_NEW_AFTER: '',
            common.SETTING_GK_EDIT: 'Ctrl+E',
            common.SETTING_GK_TOGGLE_DONE: 'Ctrl+D',
            common.SETTING_GK_SEARCH: 'Ctrl+F',
            common.SETTING_GK_TOGGLE_TRACKING: 'Ctrl+T',
            common.SETTING_GK_OPEN_MANUAL: 'F1',
            common.SETTING_GK_OPEN_FILE: '',
            common.SETTING_GK_NAMED_SEARCHES: 'F8',
            common.SETTING_GK_FOCUS_TASKS: 'F6',
            common.SETTING_GK_TOGGLE_HIDDEN: 'Ctrl+H',
            common.SETTING_GK_DELEGATE: 'Ctrl+G',
            common.SETTING_GK_TOGGLE_DARK: '',
        },
}


class Configuration:
    def __init__(self, conf):
        self.conf = conf

    def __getitem__(self, group):
        return self.conf[group]

    @property
    def sections(self):
        return self.conf.sections()

    def __contains__(self, item):
        return item in self.conf

    def get(self, group, item, default=None):
        if group in self.conf:
            return self.conf[group].get(item, default)
        else:
            return default

    def bool(self, group, item, default='n'):
        return self.get(group, item, default).lower() in ['y', 'yes', '1', 'true', 'on']

    def list(self, group, item, default='', sep=',', strip=True):
        return [e.strip() if strip else e for e in self.get(group, item, default).split(sep)]

    def number(self, group, item, default='0'):
        value = self.get(group, item, default)
        if value.isnumeric():
            return int(value)
        return None

    def time(self, group, item, default=None):
        value = self.get(group, item, None)
        try:
            return datetime.datetime.strptime(value, "%H:%M")
        except ValueError:
            return default

    def color_pair(self, group, item, default=None):
        value = self.get(group, item, default)
        if value is None or len(value) == 0:
            return None

        result = [None, None]
        for idx, colnr in enumerate(value.split(',')):
            if idx >= 2:
                break
            
            colnr = colnr.strip()

            if not colnr.isnumeric():
                # TODO: accept some colors by name
                continue

            result[idx] = int(colnr)

        return result


def get_config(args):
    conf = configparser.ConfigParser(interpolation=None)
    conf.read_dict(DEFAULT_CONFIG)
    conffile = pathlib.Path(args.config).expanduser().resolve()

    if conffile.exists() and conffile.is_file():
        conf.read([conffile])

    if common.SETTING_GROUP_INCLUDE in conf and 'path' in conf[common.SETTING_GROUP_INCLUDE]:
        path = pathlib.Path(conf[common.SETTING_GROUP_INCLUDE]['path']).expanduser().resolve()
        conf.read([path])
    
    return Configuration(conf)




import unittest
import pathlib

from pytodotxt import Task

from pter import utils
from pter.searcher import Searcher
from pter.source import Source


class FakeSource:
    def __init__(self, tasks):
        self.tasks = tasks
        self.filename = pathlib.Path('/tmp/test.txt')


class SearcherTest(unittest.TestCase):
    def setUp(self):
        self.searcher = Searcher('', False)

    def search(self, text, tasks):
        self.searcher.text = text
        self.searcher.parse()
        result = []
        source = Source(FakeSource(tasks))
        source.update_contexts_and_projects()
        self.searcher.update_sources([source])
        for task in source.tasks:
            task.todotxt = source
            if self.searcher.match(task):
                result.append(task)
        return result


class TestAfter(SearcherTest):
    def test_show_all(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 after:1')]
        results = self.search('after:', tasks)
        self.assertEqual(len(results), 2)

    def test_hide_after(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 after:1')]
        results = self.search('', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['1'])

    def test_recursion(self):
        tasks = [Task('a id:1 after:3'),
                 Task('b id:2 after:1'),
                 Task('c id:3 after:2')]
        results = self.search('', tasks)
        self.assertEqual(len(results), 0)

    def test_more_parents(self):
        tasks = [Task('a id:1'),
                 Task('b id:2'),
                 Task('c after:1,2')]
        results = self.search('', tasks)
        self.assertEqual(len(results), 2)

        results = self.search('after:1', tasks)
        self.assertEqual(len(results), 1)
        self.assertIn('c', str(results[0]))

    def test_parent_completed(self):
        tasks = [Task('x a id:1'),
                 Task('b after:1 id:test')]
        results = self.search('done:n', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['test'])

    def test_some_parents_completed(self):
        tasks = [Task('x a id:1'),
                 Task('b id:2'),
                 Task('c id:3 after:1,2')]
        results = self.search('done:n', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['2'])

    def test_some_parents_completed2(self):
        tasks = [Task('x a id:1'),
                 Task('b id:2'),
                 Task('c id:3 after:1 after:2')]
        results = self.search('done:n', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['2'])

    def test_all_parents_completed(self):
        tasks = [Task('x a id:1'),
                 Task('x b id:2'),
                 Task('c id:3 after:1,2')]
        results = self.search('done:n', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['3'])

    def test_all_parents_completed2(self):
        tasks = [Task('x a id:1'),
                 Task('x b id:2'),
                 Task('c id:3 after:1 after:2')]
        results = self.search('done:n', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['3'])


class TestIDs(SearcherTest):
    def test_id(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1')]
        results = self.search('id:1', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['1'])

    def test_id_not_there(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1')]
        results = self.search('id:3', tasks)
        self.assertEqual(len(results), 0)

    def test_ids(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1')]
        results = self.search('id:1,2', tasks)
        self.assertEqual(len(results), 2)

    def test_ids2(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1')]
        results = self.search('id:1 id:2', tasks)
        self.assertEqual(len(results), 2)

    def test_not_id(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1')]
        results = self.search('-id:1', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['2'])

    def test_not_ids(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1')]
        results = self.search('-id:1,2', tasks)
        self.assertEqual(len(results), 0)

    def test_not_ids2(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1')]
        results = self.search('-id:1 -id:2', tasks)
        self.assertEqual(len(results), 0)

    def test_has_id(self):
        tasks = [Task('a id:1'),
                 Task('b')]
        results = self.search('id:', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['1'])

    def test_has_no_id(self):
        tasks = [Task('a id:1'),
                 Task('b')]
        results = self.search('-id:', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(str(results[0]), 'b')


class TestRef(SearcherTest):
    def test_ref_not_there(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1')]
        results = self.search('ref:2', tasks)
        self.assertEqual(len(results), 0)

    def test_ref_search(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1')]
        results = self.search('ref:1', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['2'])

    def test_after_search(self):
        tasks = [Task('x a id:1'),
                 Task('t id:2 after:1')]
        results = self.search('ref:1', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['2'])

    def test_ref_multiple(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1,4')]
        results = self.search('ref:4', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['2'])

    def test_ref_multiple2(self):
        tasks = [Task('a id:1'),
                 Task('b id:2 ref:1 ref:4')]
        results = self.search('ref:4', tasks)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].attr_id, ['2'])


class SearchFilename(SearcherTest):
    def test_match(self):
        tasks = [Task('a id:1')]
        results = self.search('file:test', tasks)
        self.assertEqual(len(results), 1)

    def test_no_match(self):
        tasks = [Task('a id:1')]
        results = self.search('file:nope', tasks)
        self.assertEqual(len(results), 0)

    def test_match_not(self):
        tasks = [Task('a id:1')]
        results = self.search('not:file:test', tasks)
        self.assertEqual(len(results), 0)

    def test_no_match_not(self):
        tasks = [Task('a id:1')]
        results = self.search('not:file:nope', tasks)
        self.assertEqual(len(results), 1)




import setuptools
import pathlib

try:
    import docutils.core
    from docutils.writers import manpage
except ImportError:
    docutils = None
    manpage = None

from pter import version


with open('README.md', encoding='utf-8') as fd:
    long_description = fd.read()


with open('LICENSE', encoding='utf-8') as fd:
    licensetext = fd.read()


def compile_documentation():
    htmlfiles = []

    if docutils is None:
        return htmlfiles

    dst = pathlib.Path('./pter/docs')
    dst.mkdir(exist_ok=True)
    
    pathlib.Path('./man').mkdir(exist_ok=True)

    man_pter = None

    if None not in [docutils, manpage]:
        for fn in pathlib.Path('./doc').iterdir():
            if fn.suffix == '.rst':
                if fn.stem == 'pter':
                    man_pter = str(fn)
                dstfn = str(dst / (fn.stem + '.html'))
                docutils.core.publish_file(source_path=str(fn),
                                           destination_path=dstfn,
                                           writer_name='html')
                htmlfiles.append('docs/' + fn.stem + '.html')

    if man_pter is not None:
        docutils.core.publish_file(source_path=man_pter,
                                   destination_path='man/pter.1',
                                   writer_name='manpage')

    return htmlfiles


def collect_icons():
    icons = []
    dst = pathlib.Path('./pter/icons')

    for fn in dst.iterdir():
        if fn.is_file() and fn.suffix == '.png':
            icons.append('icons/' + fn.name)

    return icons


setuptools.setup(
    name='pter',
    version=version.__version__,
    description="Console UI to manage your todo.txt file(s).",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/vonshednob/pter",
    author="R",
    author_email="devel+pter@kakaomilchkuh.de",
    entry_points={'console_scripts': ['pter=pter.main:run'],
                  'gui_scripts': ['qpter=pter.main:run']},
    packages=['pter'],
    package_data={'pter': collect_icons() + compile_documentation()},
    data_files=[('share/man/man1', ['man/pter.1']),
                ('share/applications', ['extras/pter.desktop', 'extras/qpter.desktop']),
                ('share/doc/pter', ['extras/example.conf'])],
    install_requires=['pytodotxt>=1.0.3'],
    extras_require={'xdg': ['pyxdg']},
    python_requires='>=3.0',
    classifiers=['Development Status :: 4 - Beta',
                 'Environment :: Console :: Curses',
                 'Intended Audience :: End Users/Desktop',
                 'License :: OSI Approved :: MIT License',
                 'Natural Language :: English',
                 'Programming Language :: Python :: 3',])

